# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import sys
from typing import List, Sequence, Optional, Dict, AbstractSet, Set, Any

import click

from xcube.constants import EXTENSION_POINT_DATA_OPENERS, LOG
from xcube.constants import EXTENSION_POINT_DATA_STORES
from xcube.constants import EXTENSION_POINT_DATA_WRITERS
from xcube.util.plugin import get_extension_registry

_NO_TITLE = "<no title>"
_NO_DESCRIPTION = "<no description>"
_UNKNOWN_EXTENSION = "<unknown!>"


@click.command(name='list')
def store_list():
    """List names of data stores."""
    print(f'Data stores:')
    count = _dump_data_stores()
    print(f'{count} data store{"s" if count != 1 else ""} found.')


@click.command(name='list')
def opener_list():
    """List names of data openers."""
    print(f'Data openers:')
    count = _dump_data_openers()
    print(f'{count} data opener{"s" if count != 1 else ""} found.')


@click.command(name='list')
def writer_list():
    """List names of data writers."""
    print(f'Data writers:')
    count = _dump_data_writers()
    print(f'{count} data writer{"s" if count != 1 else ""} found.')


@click.command(name='info')
@click.argument('store_id', metavar='STORE')
@click.argument('store_params', metavar='PARAMS', nargs=-1)
@click.option('-P', '--params', 'show_params', is_flag=True, help='Show data store parameters.')
@click.option('-O', '--openers', 'show_openers', is_flag=True, help='Show available data store openers.')
@click.option('-W', '--writers', 'show_writers', is_flag=True, help='Show available data store writers.')
@click.option('-D', '--data', 'show_data_ids', is_flag=True, help='Show available data resources.')
@click.option('-j', '--json', 'use_json_format', is_flag=True, help='Output using JSON format.')
def store_info(store_id: str,
               store_params: List[str],
               show_params: bool,
               show_openers: bool,
               show_writers: bool,
               show_data_ids: bool,
               use_json_format: bool):
    """
    Show data store information.

    Dumps detailed data store information in human readable form or as JSON, when using the --json option.

    You can obtain valid STORE names using command "xcube store list".

    Note some stores require provision of parameters PARAMS
    when using one of the options --openers/-O, --writers/-W, or --data/-D.
    To find out which parameters are available use the command with just the --params/-P option first.
    """
    extension = get_extension_registry().get_extension(EXTENSION_POINT_DATA_STORES, store_id)
    from xcube.core.store import get_data_store_params_schema
    from xcube.core.store import MutableDataStore
    from xcube.core.store import DataStoreError
    params_schema = get_data_store_params_schema(store_id)
    description = extension.metadata.get('description')
    requires_store_instance = any((show_openers, show_writers, show_data_ids))
    data_store = None
    try:
        if requires_store_instance:
            data_store = _new_data_store(store_id, store_params)
    except DataStoreError as e:
        msg = f'Failed to instantiate data store {store_id!r}'
        if store_params:
            msg += f' with parameters {store_params!r}'
        msg += f': {e}'
        raise click.ClickException(msg) from e
    if use_json_format:
        d = dict()
        d['store_id'] = store_id
        if description:
            d['description'] = description
        if show_params:
            d['params_schema'] = params_schema.to_dict()
        if show_openers:
            d['opener_ids'] = data_store.get_data_opener_ids()
        if show_writers and isinstance(data_store, MutableDataStore):
            d['writer_ids'] = data_store.get_data_writer_ids()
        if show_data_ids:
            d['data_ids'] = sorted(data_store.get_data_ids())
        if show_openers:
            print(json.dumps(d, indent=2))
    else:
        print(f'\nData store description:')
        print(f'  {description or _NO_DESCRIPTION}')
        if show_params:
            print(_format_params_schema(params_schema))
        if show_openers:
            print(f'\nData openers:')
            _dump_store_openers(data_store)
        if show_writers:
            if isinstance(data_store, MutableDataStore):
                print(f'\nData writers:')
                _dump_store_writers(data_store)
            else:
                print(f'No writers available, because data store "{store_id}" is not mutable.')
        if show_data_ids:
            print(f'\nData resources:')
            count = _dump_store_data_ids(data_store)
            print(f'{count} data resource{"s" if count != 1 else ""} found.')


@click.command(name='data')
@click.argument('store_id', metavar='STORE')
@click.argument('data_id', metavar='DATA')
@click.argument('store_params', metavar='PARAMS', nargs=-1)
def store_data(store_id: str, data_id: str, store_params: List[str]):
    """
    Show data resource information.

    Show the data descriptor for data resource DATA in data store STORE.
    Note some stores require provision of store parameters PARAMS.
    Use "xcube io store info STORE -P" command to find out which parameters are available/required.
    """
    data_store = _new_data_store(store_id, store_params)
    data_descriptor = data_store.describe_data(data_id)
    print(f'Descriptor for data resource "{data_id}" in data store "{store_id}":')
    print(json.dumps(data_descriptor.to_dict(), indent=2))


@click.command(name='info')
@click.argument('opener_id', metavar='OPENER')
def opener_info(opener_id: str):
    """
    Show data opener information.
    You can obtain valid OPENER names using command "xcube io opener list".
    """
    extension = get_extension_registry().get_extension(EXTENSION_POINT_DATA_OPENERS, opener_id)
    description = extension.metadata.get('description')
    if description:
        print(description)
    from xcube.core.store import new_data_opener
    opener_ = new_data_opener(opener_id)
    params_schema = opener_.get_open_data_params_schema()
    print(_format_params_schema(params_schema))


@click.command(name='info')
@click.argument('writer_id', metavar='WRITER')
def writer_info(writer_id: str):
    """
    Show data opener information.
    You can obtain valid WRITER names using command "xcube io writer list".
    """
    extension = get_extension_registry().get_extension(EXTENSION_POINT_DATA_WRITERS, writer_id)
    description = extension.metadata.get('description')
    if description:
        print(description)
    from xcube.core.store import new_data_writer
    writer_ = new_data_writer(writer_id)
    params_schema = writer_.get_write_data_params_schema()
    print(_format_params_schema(params_schema))


_DEFAULT_DUMP_OUTPUT = 'store-dump.json'
_SHORT_INCLUDE = ','.join(['store.store_instance_id',
                           'store.title',
                           'data.data_id',
                           'data.bbox',
                           'data.spatial_res',
                           'data.time_range',
                           'data.time_period',
                           'var.name',
                           'var.dtype',
                           'var.dims'])


@click.command(name='dump')
@click.option('-o', '--output', 'output_file_path', metavar='OUTPUT', default=None,
              help=f'Output filename. Output will be written as JSON.'
                   f' Defaults to "{_DEFAULT_DUMP_OUTPUT}".')
@click.option('-c', '--config', 'config_file_path', metavar='CONFIG',
              help='Store configuration filename. May use JSON or YAML format.')
@click.option('-t', '--type', 'data_type', metavar='TYPE',
              help='Data type. If given, only data resources that satisfy the '
                   'data type are listed. E.g. "dataset" or "geodataframe"')
@click.option('-S', '--short', 'short_form', is_flag=True,
              help=f'Short form. Forces option "--includes={_SHORT_INCLUDE}".')
@click.option('-I', '--includes', 'include_props', metavar='INCLUDE_LIST',
              help='Comma-separated list of properties to be included'
                   ' from stores (prefix "store."),'
                   ' data resources (prefix "data.") of stores,'
                   ' and variables (prefix "var.") of data resources.')
@click.option('-E', '--excludes', 'exclude_props', metavar='EXCLUDE_LIST',
              help='Comma-separated list of properties to be excluded'
                   ' from stores (prefix "store."),'
                   ' data resources (prefix "data.") of stores,'
                   ' and variables (prefix "var.") of data resources.')
@click.option('--csv', 'csv_format', is_flag=True,
              help=f'Use CSV output format.')
@click.option('--yaml', 'yaml_format', is_flag=True,
              help=f'Use YAML output format.')
@click.option('--json', 'json_format', is_flag=True,
              help=f'Use JSON output format (the default).')
def dump(output_file_path: Optional[str],
         config_file_path: Optional[str],
         data_type: Optional[str],
         short_form: bool,
         include_props: str,
         exclude_props: str,
         csv_format: bool,
         yaml_format: bool,
         json_format: bool):
    """
    Dump metadata of given data stores.

    Dumps data store metadata and metadata for a store's data resources
    for given data stores  into a JSON file.
    Data stores may be selected and configured by a configuration file CONFIG,
    which may have JSON or YAML format.
    For example, this YAML configuration configures a single directory data store:

    \b
    this_dir:
        title: Current Dir
        description: A store that represents my current directory
        store_id: "directory"
        store_params:
            base_dir: "."

    """
    from xcube.core.store import DataStoreConfig
    from xcube.core.store import DataStorePool
    import yaml
    import json
    import os.path

    if csv_format:
        output_format = 'csv'
        ext = '.csv'
    elif yaml_format:
        output_format = 'yaml'
        ext = '.yml'
    elif json_format:
        output_format = 'json'
        ext = '.json'
    elif output_file_path is not None:
        path_no_ext, ext = os.path.splitext(output_file_path)
        if ext in ('.csv', '.txt'):
            output_format = 'csv'
        elif ext in ('.yaml', '.yml'):
            output_format = 'yaml'
        else:
            output_format = 'json'
    else:
        output_format = 'json'
        ext = '.json'

    if output_file_path is None:
        path_no_ext, _ = os.path.splitext(_DEFAULT_DUMP_OUTPUT)
        output_file_path = path_no_ext + ext

    include_props = _parse_props(include_props) if include_props else None
    exclude_props = _parse_props(exclude_props) if exclude_props else None

    if short_form:
        short_include_props = _parse_props(_SHORT_INCLUDE)
        include_props = include_props or {}
        for data_key in ('store', 'data', 'var'):
            include_props[data_key] = include_props.get(data_key, set()).union(short_include_props[data_key])

    if config_file_path:
        store_pool = DataStorePool.from_file(config_file_path)
    else:
        extensions = get_extension_registry().find_extensions(EXTENSION_POINT_DATA_STORES)
        store_configs = {extension.name: DataStoreConfig(extension.name,
                                                         title=extension.metadata.get('title'),
                                                         description=extension.metadata.get('description'))
                         for extension in extensions
                         if extension.name not in ('memory', 'directory', 's3')}
        store_pool = DataStorePool(store_configs)

    dump_data = _get_store_data_var_tuples(store_pool, data_type, include_props, exclude_props)

    if output_format == 'csv':
        column_names = None
        column_names_set = None
        rows = []
        for store_dict, data_dict, var_dict in dump_data:
            if store_dict is None:
                break
            row = {}
            row.update({'store.' + k: v for k, v in store_dict.items()})
            row.update({'data.' + k: v for k, v in data_dict.items()})
            row.update({'var.' + k: v for k, v in var_dict.items()})
            rows.append(row)
            if column_names_set is None:
                column_names = list(row.keys())
                column_names_set = set(column_names)
            else:
                for k in row.keys():
                    if k not in column_names_set:
                        column_names.append(k)
                        column_names_set.add(k)

        def format_cell_value(value: Any) -> str:
            return str(value) if value is not None else ''

        sep = '\t'
        with open(output_file_path, 'w') as fp:
            if column_names:
                fp.write(sep.join(column_names) + '\n')
                for row in rows:
                    fp.write(sep.join(map(format_cell_value,
                                          tuple(row.get(k)
                                                for k in
                                                column_names))) + '\n')

        LOG.info(f'Dumped {len(rows)} store entry/ies'
                 f' to {output_file_path}.')

    else:
        last_store_dict = None
        last_data_dict = None
        vars_list = []
        data_list = []
        store_list = []
        for store_dict, data_dict, var_dict in dump_data:
            if data_dict is not last_data_dict or data_dict is None:
                if last_data_dict is not None:
                    last_data_dict['data_vars'] = vars_list
                    vars_list = []
                    data_list.append(last_data_dict)
                last_data_dict = data_dict
            if store_dict is not last_store_dict or store_dict is None:
                if last_store_dict is not None:
                    last_store_dict['data'] = data_list
                    data_list = []
                    store_list.append(last_store_dict)
                last_store_dict = store_dict
            if var_dict:
                vars_list.append(var_dict)

        with open(output_file_path, 'w') as fp:
            if output_format == 'json':
                from xcube.util.jsonencoder import NumpyJSONEncoder
                json.dump(dict(stores=store_list), fp, indent=2,
                          cls=NumpyJSONEncoder)
            else:
                yaml.dump(dict(stores=store_list), fp, indent=2)

        LOG.info(f'Dumped entries of {len(store_list)} store(s)'
                 f' to {output_file_path}.')


def _get_store_data_var_tuples(store_pool, data_type, include_props, exclude_props):
    import time

    for store_instance_id in store_pool.store_instance_ids:
        t0 = time.perf_counter()
        print(f'Generating entries for store "{store_instance_id}"...')

        store_config = store_pool.get_store_config(store_instance_id)
        store_dict = dict(store_instance_id=store_instance_id,
                          store_id=store_instance_id,
                          title=store_config.title,
                          description=store_config.description,
                          data_type=data_type,
                          data=[])
        store_dict = _filter_dict(store_dict, 'store', include_props, exclude_props)

        if 'data' not in store_dict:
            yield store_dict, {}, {}
        else:
            del store_dict['data']

            try:
                store_instance = store_pool.get_store(store_instance_id)
            except BaseException as error:
                LOG.error(f'Cannot open store'
                          f' "{store_instance_id}": {error}')
                continue

            try:
                data_descriptors = store_instance.search_data(data_type=data_type)
            except BaseException as error:
                LOG.error(f'Cannot search store'
                          f' "{store_instance_id}": {error}')
                continue

            for data_descriptor in data_descriptors:
                print(f'Processing data resource "{data_descriptor.data_id}"...')
                data_dict = data_descriptor.to_dict()
                data_dict = _filter_dict(data_dict, 'data', include_props, exclude_props)
                if 'data_vars' not in data_dict:
                    yield store_dict, data_dict, {}
                else:
                    var_dicts = data_dict.pop('data_vars')
                    for var_name, var_dict in var_dicts.items():
                        var_dict['name'] = var_name
                        var_dict = _filter_dict(var_dict, 'var', include_props, exclude_props)
                        yield store_dict, data_dict, var_dict

        print(f'Done generating entries for store "{store_instance_id}" after '
              + '{:.2f} seconds'.format(time.perf_counter() - t0))

    # yield Terminator
    yield None, None, None


def _filter_dict(data: Dict[str, Any],
                 selector: str,
                 include_props: Dict[str, Set[str]] = None,
                 exclude_props: Dict[str, Set[str]] = None) -> Dict[str, Any]:
    includes = (include_props.get(selector) or None) if include_props is not None else None
    excludes = (exclude_props.get(selector) or None) if exclude_props is not None else None
    if includes is None and excludes is None:
        return data
    return {key: value
            for key, value in data.items()
            if (includes is None or key in includes)
            and (excludes is None or key not in excludes)}


def _parse_props(props: str) -> Dict[str, AbstractSet]:
    parsed_props = dict(store=set(), data=set(), var=set())
    for p in props.split(','):
        try:
            prefix, name = p.strip().split('.')
            parsed_props[prefix].add(name)
        except (ValueError, KeyError):
            raise click.ClickException(f'Invalid include/exclude property: {p}')
    if parsed_props['var']:
        parsed_props['data'].add('data_vars')
    if parsed_props['data']:
        parsed_props['store'].add('data')
    return parsed_props


@click.group()
def store():
    """
    Tools for xcube's data stores.
    """
    pass


@click.group()
def opener():
    """
    Tools for xcube's data openers.
    """
    pass


@click.group()
def writer():
    """
    Tools for xcube's data writers.
    """
    pass


store.add_command(store_list)
store.add_command(store_info)
store.add_command(store_data)
opener.add_command(opener_list)
opener.add_command(opener_info)
writer.add_command(writer_list)
writer.add_command(writer_info)


@click.group(hidden=True)
def io():
    """
    Tools for xcube's generic I/O system.
    """
    pass


io.add_command(store)
io.add_command(opener)
io.add_command(writer)
io.add_command(dump)


# from xcube.util.jsonschema import JsonObjectSchema


# noinspection PyUnresolvedReferences
def _format_params_schema(params_schema: 'xcube.util.jsonschema.JsonObjectSchema') -> str:
    text = []
    if params_schema.properties:
        text.append('Parameters:')
        for param_name, param_schema in params_schema.properties.items():
            text.append(f'{"* " if param_name in params_schema.required else "  "}'
                        f'{param_name:>24s}  {_format_param_schema(param_schema)}')
    else:
        text.append(f'No parameters required.')
    return '\n'.join(text)


# noinspection PyUnresolvedReferences
def _format_required_params_schema(params_schema: 'xcube.util.jsonschema.JsonObjectSchema') -> str:
    text = ['Required parameters:']
    for param_name, param_schema in params_schema.properties.items():
        if param_name in params_schema.required:
            text.append(f'  {param_name:>24s}  {_format_param_schema(param_schema)}')
    return '\n'.join(text)


# noinspection PyUnresolvedReferences
def _format_param_schema(param_schema: 'xcube.util.jsonschema.JsonSchema'):
    from xcube.util.undefined import UNDEFINED
    param_info = []
    if param_schema.title:
        param_info.append(param_schema.title + ('' if param_schema.title.endswith('.') else '.'))
    if param_schema.description:
        param_info.append(param_schema.description + ('' if param_schema.description.endswith('.') else '.'))
    if param_schema.enum:
        param_info.append(f'Must be one of {", ".join(map(json.dumps, param_schema.enum))}.')
    if param_schema.const != UNDEFINED:
        param_info.append(f'Must be {json.dumps(param_schema.const)}.')
    if param_schema.default != UNDEFINED:
        param_info.append(f'Defaults to {json.dumps(param_schema.default)}.')
    param_info_text = ' ' + " ".join(param_info) if param_info else ''
    return f'({param_schema.type}){param_info_text}'


def _dump_data_stores() -> int:
    return _dump_extensions(EXTENSION_POINT_DATA_STORES)


def _dump_data_openers() -> int:
    return _dump_extensions(EXTENSION_POINT_DATA_OPENERS)


def _dump_data_writers() -> int:
    return _dump_extensions(EXTENSION_POINT_DATA_WRITERS)


def _dump_extensions(point: str) -> int:
    count = 0
    for extension in get_extension_registry().find_extensions(point):
        print(f'  {extension.name:>24s}  {extension.metadata.get("description", "<no description>")}')
        count += 1
    return count


# noinspection PyUnresolvedReferences
def _dump_store_openers(data_store: 'xcube.core.store.DataStore', data_id: str = None) -> int:
    return _dump_named_extensions(EXTENSION_POINT_DATA_OPENERS, data_store.get_data_opener_ids(data_id=data_id))


# noinspection PyUnresolvedReferences
def _dump_store_writers(data_store: 'xcube.core.store.DataStore') -> int:
    return _dump_named_extensions(EXTENSION_POINT_DATA_WRITERS, data_store.get_data_writer_ids())


# noinspection PyUnresolvedReferences
def _dump_store_data_ids(data_store: 'xcube.core.store.DataStore') -> int:
    count = 0
    for data_id, data_attrs in sorted(data_store.get_data_ids(include_attrs=['title'])):
        print(f'  {data_id:>32s}  {data_attrs.get("title") or _NO_TITLE}')
        count += 1
    return count


def _dump_named_extensions(point: str, names: Sequence[str]) -> int:
    count = 0
    for name in names:
        extension = get_extension_registry().get_extension(point, name)
        if extension:
            print(f'  {name:>24s}  {extension.metadata.get("description", _NO_DESCRIPTION)}')
        else:
            print(f'  {name:>24s}  {_UNKNOWN_EXTENSION}')
        count += 1
    return count


# noinspection PyUnresolvedReferences
def _dump_data_resources(data_store: 'xcube.core.store.DataStore') -> int:
    count = 0
    for data_id, title in data_store.get_data_ids():
        print(f'  {data_id:<32s}  {title or _NO_TITLE}')
        count += 1
    return count


# noinspection PyUnresolvedReferences
def _new_data_store(store_id: str, store_params: List[str]) -> 'xcube.core.store.DataStore':
    from xcube.core.store import get_data_store_params_schema
    from xcube.core.store import new_data_store
    store_params_dict = dict()
    if store_params:
        for p_assignment in store_params:
            p_name_value = p_assignment.split('=', maxsplit=2)
            p_name = p_name_value[0].strip()
            if not p_name:
                raise click.ClickException(f'Invalid parameter assignment: {p_assignment}')
            if len(p_name_value) == 2:
                # Passed as name=value
                p_value = p_name_value[1].strip()
                try:
                    p_value = json.loads(p_value)
                except json.decoder.JSONDecodeError:
                    pass
            else:
                # Passed name as flag
                p_value = True
            store_params_dict[p_name] = p_value
    return new_data_store(store_id, **store_params_dict)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import fnmatch
import functools
import io
import json
from typing import Dict, Tuple, List, Set, Optional, Any, Callable, Mapping

import matplotlib.colorbar
import matplotlib.colors
import matplotlib.figure
import numpy as np
import pyproj
import xarray as xr

from xcube.constants import LOG
from xcube.core.geom import get_dataset_bounds
from xcube.core.geom import get_dataset_geometry
from xcube.core.normalize import DatasetIsNotACubeError
from xcube.core.store import DataStoreError
from xcube.core.tilingscheme import TilingScheme
from xcube.core.timecoord import timestamp_to_iso_string
from xcube.server.api import ApiError
from xcube.constants import CRS_CRS84
from .authutil import READ_ALL_DATASETS_SCOPE
from .authutil import READ_ALL_VARIABLES_SCOPE
from .authutil import assert_scopes
from .authutil import check_scopes
from .context import DatasetConfig
from .context import DatasetsContext
from ..places.controllers import GeoJsonFeatureCollection
from ..places.controllers import find_places


def find_dataset_places(ctx: DatasetsContext,
                        place_group_id: str,
                        ds_id: str,
                        base_url: str,
                        query_expr: Any = None,
                        comb_op: str = "and") -> GeoJsonFeatureCollection:
    dataset = ctx.get_dataset(ds_id)
    query_geometry = get_dataset_geometry(dataset)
    return find_places(ctx.places_ctx,
                       place_group_id,
                       base_url,
                       query_geometry=query_geometry,
                       query_expr=query_expr,
                       comb_op=comb_op)


def get_datasets(ctx: DatasetsContext,
                 details: bool = False,
                 point: Optional[Tuple[float, float]] = None,
                 base_url: Optional[str] = None,
                 granted_scopes: Optional[Set[str]] = None) -> Dict:
    can_authenticate = ctx.can_authenticate
    # If True, we can shorten scope checking
    if granted_scopes is None:
        can_read_all_datasets = False
    else:
        can_read_all_datasets = READ_ALL_DATASETS_SCOPE in granted_scopes

    LOG.info(f'Collecting datasets for granted scopes {granted_scopes!r}')

    dataset_configs = list(ctx.get_dataset_configs())

    dataset_dicts = list()
    for dataset_config in dataset_configs:

        ds_id = dataset_config['Identifier']

        if dataset_config.get('Hidden'):
            continue

        if can_authenticate \
                and not can_read_all_datasets \
                and not _allow_dataset(ctx,
                                       dataset_config,
                                       granted_scopes,
                                       check_scopes):
            LOG.info(
                f'Rejected dataset {ds_id!r} due to missing permission'
            )
            continue

        dataset_dict = dict(id=ds_id)

        dataset_dict['title'] = ds_id
        if 'Title' in dataset_config:
            ds_title = dataset_config['Title']
            if ds_title and isinstance(ds_title, str):
                dataset_dict['title'] = ds_title

        ds_bbox = dataset_config.get('BoundingBox')
        if ds_bbox is not None:
            # Note, it has already been validated that ds_bbox is valid
            dataset_dict['bbox'] = ds_bbox

        LOG.info(f'Collected dataset {ds_id}')
        dataset_dicts.append(dataset_dict)

    # Important note:
    # the "point" parameter is used by
    # the CyanoAlert app only

    if details or point:
        filtered_dataset_dicts = []
        for dataset_dict in dataset_dicts:
            ds_id = dataset_dict["id"]
            try:
                if point:
                    ds = ctx.get_dataset(ds_id)
                    if "bbox" not in dataset_dict:
                        dataset_dict["bbox"] = list(get_dataset_bounds(ds))
                if details:
                    LOG.info(f'Loading details for dataset {ds_id}')
                    dataset_dict.update(
                        get_dataset(ctx, ds_id,
                                    base_url,
                                    granted_scopes=granted_scopes)
                    )
                filtered_dataset_dicts.append(dataset_dict)
            except Exception as e:
                LOG.warning(f'Skipping dataset {ds_id}: {e}', exc_info=True)
        dataset_dicts = filtered_dataset_dicts

    if point:
        is_point_in_dataset_bbox = functools.partial(
            _is_point_in_dataset_bbox, point
        )
        # noinspection PyTypeChecker
        dataset_dicts = list(filter(is_point_in_dataset_bbox, dataset_dicts))

    if not dataset_dicts:
        LOG.warn(f'No datasets provided for current user.')

    return dict(datasets=dataset_dicts)


def get_dataset(ctx: DatasetsContext,
                ds_id: str,
                base_url: Optional[str] = None,
                granted_scopes: Optional[Set[str]] = None) -> Dict:
    can_authenticate = ctx.can_authenticate
    # If True, we can shorten scope checking
    if granted_scopes is None:
        can_read_all_datasets = False
        can_read_all_variables = False
    else:
        can_read_all_datasets = READ_ALL_DATASETS_SCOPE in granted_scopes
        can_read_all_variables = READ_ALL_VARIABLES_SCOPE in granted_scopes

    dataset_config = ctx.get_dataset_config(ds_id)
    ds_id = dataset_config['Identifier']

    if can_authenticate and not can_read_all_datasets:
        _allow_dataset(ctx, dataset_config, granted_scopes, assert_scopes)

    try:
        ml_ds = ctx.get_ml_dataset(ds_id)
    except (ValueError, DataStoreError) as e:
        raise DatasetIsNotACubeError(f'could not open dataset: {e}') from e

    ds = ml_ds.get_dataset(0)

    try:
        ts_ds = ctx.get_time_series_dataset(ds_id)
    except (ValueError, DataStoreError):
        ts_ds = None

    x_name, y_name = ml_ds.grid_mapping.xy_dim_names

    ds_title = dataset_config.get('Title',
                                  ds.attrs.get('title',
                                               ds.attrs.get('name',
                                                            ds_id)))
    dataset_dict = dict(id=ds_id, title=ds_title)

    crs = ml_ds.grid_mapping.crs
    transformer = pyproj.Transformer.from_crs(crs, CRS_CRS84, always_xy=True)
    dataset_bounds = get_dataset_bounds(ds)

    ds_bbox = dataset_config.get('BoundingBox')
    if ds_bbox is not None:
        # Note, JSON-Schema already verified that ds_bbox is valid.
        # 'BoundingBox' is always given in geographical coordinates.
        ds_bbox = list(ds_bbox)
    else:
        x1, y1, x2, y2 = dataset_bounds
        if not crs.is_geographic:
            (x1, x2), (y1, y2) = transformer.transform((x1, x2), (y1, y2))
        ds_bbox = [x1, y1, x2, y2]

    dataset_dict['bbox'] = list(ds_bbox)
    dataset_dict['geometry'] = get_bbox_geometry(dataset_bounds, transformer)
    dataset_dict['spatialRef'] = crs.to_string()

    variable_dicts = []
    dim_names = set()

    tiling_scheme = ml_ds.derive_tiling_scheme(TilingScheme.GEOGRAPHIC)
    LOG.debug('Tile level range for dataset %s: %d to %d',
              ds_id,
              tiling_scheme.min_level,
              tiling_scheme.max_level)

    spatial_var_names = [str(var_name)
                         for var_name, var in ds.data_vars.items()
                         if (len(var.dims) >= 2
                             and var.dims[-2] == y_name
                             and var.dims[-1] == x_name)]

    var_name_patterns = dataset_config.get('Variables')
    if var_name_patterns:
        spatial_var_names = filter_variable_names(spatial_var_names,
                                                  var_name_patterns)
        if not spatial_var_names:
            LOG.warning(f'No variable matched any of the patterns given'
                        f' in the "Variables" filter.'
                        f' You may specify a wildcard "*" as last item.')

    for var_name in spatial_var_names:
        var = ds.data_vars[var_name]

        if can_authenticate \
                and not can_read_all_variables \
                and not _allow_variable(ctx,
                                        dataset_config,
                                        var_name,
                                        granted_scopes):
            continue

        variable_dict = dict(
            id=f'{ds_id}.{var_name}',
            name=var_name,
            dims=list(var.dims),
            shape=list(var.shape),
            dtype=str(var.dtype),
            units=var.attrs.get('units', ''),
            title=var.attrs.get('title',
                                var.attrs.get('long_name',
                                              var_name)),
            timeChunkSize=get_time_chunk_size(ts_ds, var_name, ds_id)
        )

        tile_url = _get_dataset_tile_url2(ctx, ds_id, var_name, base_url)
        variable_dict["tileUrl"] = tile_url
        LOG.debug('Tile URL for variable %s: %s',
                  var_name, tile_url)

        variable_dict["tileLevelMin"] = tiling_scheme.min_level
        variable_dict["tileLevelMax"] = tiling_scheme.max_level

        cmap_name, (cmap_vmin, cmap_vmax) = ctx.get_color_mapping(ds_id,
                                                                  var_name)
        variable_dict["colorBarName"] = cmap_name
        variable_dict["colorBarMin"] = cmap_vmin
        variable_dict["colorBarMax"] = cmap_vmax

        if hasattr(var.data, '_repr_html_'):
            # noinspection PyProtectedMember
            variable_dict["htmlRepr"] = var.data._repr_html_()

        variable_dict["attrs"] = {
            key: ("NaN"
                  if isinstance(value, float) and np.isnan(value)
                  else value)
            for key, value in var.attrs.items()
        }

        variable_dicts.append(variable_dict)
        for dim_name in var.dims:
            dim_names.add(dim_name)

    dataset_dict["variables"] = variable_dicts

    if not variable_dicts:
        if not ds.data_vars:
            message = f'Dataset {ds_id!r} has no variables'
        else:
            message = f'Dataset {ds_id!r} has no published variables'
        raise DatasetIsNotACubeError(message)

    rgb_var_names, rgb_norm_ranges = ctx.get_rgb_color_mapping(ds_id)
    if any(rgb_var_names):
        rgb_tile_url = _get_dataset_tile_url2(ctx, ds_id, 'rgb', base_url)
        rgb_schema = {
            'varNames': rgb_var_names,
            'normRanges': rgb_norm_ranges,
            'tileUrl': rgb_tile_url,
            'tileLevelMin': tiling_scheme.min_level,
            'tileLevelMax': tiling_scheme.max_level,
        }

        dataset_dict["rgbSchema"] = rgb_schema

    dataset_dict["dimensions"] = [
        get_dataset_coordinates(ctx, ds_id, str(dim_name))
        for dim_name in dim_names
    ]

    dataset_dict["attrs"] = {
        key: ds.attrs[key]
        for key in sorted(list(map(str, ds.attrs.keys())))
    }

    dataset_attributions = dataset_config.get(
        'Attribution',
        ctx.config.get('DatasetAttribution')
    )
    if dataset_attributions is not None:
        if isinstance(dataset_attributions, str):
            dataset_attributions = [dataset_attributions]
        dataset_dict['attributions'] = dataset_attributions

    place_groups = ctx.get_dataset_place_groups(ds_id, base_url)
    if place_groups:
        dataset_dict["placeGroups"] = _filter_place_groups(place_groups,
                                                           del_features=True)

    return dataset_dict


def filter_variable_names(var_names: List[str],
                          var_name_patterns: List[str]) -> List[str]:
    filtered_var_names = []
    filtered_var_names_set = set()
    for var_name_pattern in var_name_patterns:
        for var_name in var_names:
            if var_name in filtered_var_names_set:
                continue
            if var_name == var_name_pattern \
                    or fnmatch.fnmatch(var_name, var_name_pattern):
                filtered_var_names.append(var_name)
                filtered_var_names_set.add(var_name)
    return filtered_var_names


def get_bbox_geometry(dataset_bounds: Tuple[float, float, float, float],
                      transformer: pyproj.Transformer,
                      n: int = 6):
    x1, y1, x2, y2 = dataset_bounds
    x_coords = []
    y_coords = []
    x_coords.extend(np.full(n - 1, x1))
    y_coords.extend(np.linspace(y1, y2, n)[: n - 1])
    x_coords.extend(np.linspace(x1, x2, n)[: n - 1])
    y_coords.extend(np.full(n - 1, y2))
    x_coords.extend(np.full(n - 1, x2))
    y_coords.extend(np.linspace(y2, y1, n)[: n - 1])
    x_coords.extend(np.linspace(x2, x1, n)[: n - 1])
    y_coords.extend(np.full(n - 1, y1))
    x_coords, y_coords = transformer.transform(x_coords, y_coords)
    coordinates = list(map(list, zip(map(float, x_coords),
                                     map(float, y_coords))))
    coordinates.append(coordinates[0])
    geometry = dict(type="Polygon", coordinates=[coordinates])
    return geometry


def get_time_chunk_size(ts_ds: Optional[xr.Dataset],
                        var_name: str,
                        ds_id: str) -> Optional[int]:
    """
    Get the time chunk size for variable *var_name*
    in time-chunked dataset *ts_ds*.

    Internal function.

    :param ts_ds: time-chunked dataset
    :param var_name: variable name
    :param ds_id: original dataset identifier
    :return: the time chunk size (integer) or None
    """
    if ts_ds is not None:
        ts_var: Optional[xr.DataArray] = ts_ds.get(var_name)
        if ts_var is not None:
            chunks = ts_var.chunks
            if chunks is None:
                LOG.warning(f'variable {var_name!r}'
                            f' in time-chunked dataset {ds_id!r}'
                            f' is not chunked')
                return None
            try:
                time_index = ts_var.dims.index('time')
                time_chunks = chunks[time_index]
            except ValueError:
                time_chunks = None
            if not time_chunks:
                LOG.warning(f'no chunks found'
                            f' for dimension \'time\''
                            f' of variable {var_name!r}'
                            f' in time-chunked dataset {ds_id!r}')
                return None
            if len(time_chunks) == 1:
                return time_chunks[0]
            return max(*time_chunks)
        else:
            LOG.warning(f'variable {var_name!r} not'
                        f' found in time-chunked dataset {ds_id!r}')
    return None


def _allow_dataset(
        ctx: DatasetsContext,
        dataset_config: DatasetConfig,
        granted_scopes: Optional[Set[str]],
        function: Callable[[Set, Optional[Set], bool], Any]
) -> Any:
    required_scopes = ctx.get_required_dataset_scopes(
        dataset_config
    )
    # noinspection PyArgumentList
    return function(required_scopes,
                    granted_scopes,
                    is_substitute=_is_substitute(dataset_config))


def _allow_variable(
        ctx: DatasetsContext,
        dataset_config: DatasetConfig,
        var_name: str,
        granted_scopes: Optional[Set[str]]
) -> bool:
    required_scopes = ctx.get_required_variable_scopes(
        dataset_config, var_name
    )
    # noinspection PyArgumentList
    return check_scopes(required_scopes,
                        granted_scopes,
                        is_substitute=_is_substitute(dataset_config))


def _is_substitute(dataset_config: DatasetConfig) -> bool:
    return dataset_config \
        .get('AccessControl', {}) \
        .get('IsSubstitute', False)


def get_dataset_place_groups(ctx: DatasetsContext,
                             ds_id: str,
                             base_url: str) -> List[GeoJsonFeatureCollection]:
    # Do not load or return features, just place group (metadata).
    place_groups = ctx.get_dataset_place_groups(ds_id, base_url,
                                                load_features=False)
    return _filter_place_groups(place_groups, del_features=True)


def get_dataset_place_group(ctx: DatasetsContext,
                            ds_id: str,
                            place_group_id: str,
                            base_url: str) -> GeoJsonFeatureCollection:
    # Load and return features for specific place group.
    place_group = ctx.get_dataset_place_group(ds_id, place_group_id, base_url,
                                              load_features=True)
    return _filter_place_group(place_group, del_features=False)


def get_dataset_coordinates(ctx: DatasetsContext,
                            ds_id: str,
                            dim_name: str) -> Dict:
    ds, var = ctx.get_dataset_and_coord_variable(ds_id, dim_name)
    if np.issubdtype(var.dtype, np.floating):
        values = list(map(float, var.values))
    elif np.issubdtype(var.dtype, np.integer):
        values = list(map(int, var.values))
    elif len(var) == 1:
        values = [timestamp_to_iso_string(var.values[0], round_fn="floor")]
    else:
        # see https://github.com/dcs4cop/xcube-viewer/issues/289
        assert len(var) > 1, "Dimension length must be greater than 0."
        values = [timestamp_to_iso_string(var.values[0], round_fn="floor")] +\
                 list(map(timestamp_to_iso_string, var.values[1:-1])) +\
                 [timestamp_to_iso_string(var.values[-1], round_fn="ceil")]
    return dict(name=dim_name,
                size=len(values),
                dtype=str(var.dtype),
                coordinates=values)


# noinspection PyUnusedLocal
def get_color_bars(ctx: DatasetsContext,
                   mime_type: str) -> str:
    cmaps = ctx.colormap_registry.to_json()
    if mime_type == 'application/json':
        return json.dumps(cmaps, indent=2)
    elif mime_type == 'text/html':
        html_head = '<!DOCTYPE html>\n' + \
                    '<html lang="en">\n' + \
                    '<head>' + \
                    '<meta charset="UTF-8">' + \
                    '<title>xcube Server Colormaps</title>' + \
                    '</head>\n' + \
                    '<body style="padding: 0.2em; font-family: sans-serif">\n'
        html_body = ''
        html_foot = '</body>\n' \
                    '</html>\n'
        for cmap_cat, cmap_desc, cmap_bars in cmaps:
            html_body += '    <h2>%s</h2>\n' % cmap_cat
            html_body += '    <p>%s</p>\n' % cmap_desc
            html_body += '    <table style=border: 0">\n'
            for cmap_bar in cmap_bars:
                cmap_name, cmap_data = cmap_bar
                cmap_image = f'<img src="data:image/png;base64,{cmap_data}"' \
                             f' width="100%%"' \
                             f' height="24"/>'
                name_cell = f'<td style="width: 5em">' \
                            f'<code>{cmap_name}</code>' \
                            f'</td>'
                image_cell = f'<td style="width: 40em">{cmap_image}</td>'
                row = f'<tr>{name_cell}{image_cell}</tr>\n'
                html_body += f'        {row}\n'
            html_body += '    </table>\n'
        return html_head + html_body + html_foot
    raise ApiError.BadRequest(
        f'Format {mime_type!r} not supported for colormaps'
    )


def _is_point_in_dataset_bbox(point: Tuple[float, float], dataset_dict: Dict):
    if 'bbox' not in dataset_dict:
        return False
    x, y = point
    x_min, y_min, x_max, y_max = dataset_dict['bbox']
    if not (y_min <= y <= y_max):
        return False
    if x_min < x_max:
        return x_min <= x <= x_max
    else:
        # Bounding box crosses anti-meridian
        return x_min <= x <= 180.0 or -180.0 <= x <= x_max


def _filter_place_group(place_group: Dict,
                        del_features: bool = False) -> Dict:
    place_group = dict(place_group)
    del place_group['sourcePaths']
    del place_group['sourceEncoding']
    if del_features:
        del place_group['features']
    return place_group


def _filter_place_groups(place_groups, del_features: bool = False) \
        -> List[Dict]:
    if del_features:
        def __filter_place_group(place_group):
            return _filter_place_group(place_group, del_features=True)
    else:
        def __filter_place_group(place_group):
            return _filter_place_group(place_group, del_features=False)

    return list(map(__filter_place_group, place_groups))


def _get_dataset_tile_url2(ctx: DatasetsContext,
                           ds_id: str,
                           var_name: str,
                           base_url: str):
    import urllib.parse
    return ctx.get_service_url(base_url,
                               'datasets',
                               urllib.parse.quote_plus(ds_id),
                               'vars',
                               urllib.parse.quote_plus(var_name),
                               'tiles2',
                               '{z}/{y}/{x}')


def get_legend(ctx: DatasetsContext,
               ds_id: str,
               var_name: str,
               params: Mapping[str, str]):
    default_cmap_cbar, (default_cmap_vmin, default_cmap_vmax) = \
        ctx.get_color_mapping(ds_id, var_name)

    try:
        cmap_name = params.get('cbar', params.get('cmap', default_cmap_cbar))
        cmap_vmin = float(params.get('vmin', str(default_cmap_vmin)))
        cmap_vmax = float(params.get('vmax', str(default_cmap_vmax)))
        cmap_w = int(params.get('width', '256'))
        cmap_h = int(params.get('height', '16'))
    except (ValueError, TypeError):
        raise ApiError.BadRequest("Invalid color legend parameter(s)")

    cmap_name, cmap = ctx.colormap_registry.get_cmap(cmap_name)
    colormap = ctx.colormap_registry.colormaps.get(cmap_name)
    assert colormap is not None

    norm = colormap.norm
    if norm is None:
        norm = matplotlib.colors.Normalize(vmin=cmap_vmin, vmax=cmap_vmax)

    try:
        _, cmap = ctx.colormap_registry.get_cmap(cmap_name)
    except ValueError:
        raise ApiError.NotFound(f"Colormap {cmap_name!r} not found")

    fig = matplotlib.figure.Figure(figsize=(cmap_w, cmap_h))
    ax1 = fig.add_subplot(1, 1, 1)
    image_legend = matplotlib.colorbar.ColorbarBase(ax1,
                                                    format='%.1f',
                                                    ticks=None,
                                                    cmap=cmap,
                                                    norm=norm,
                                                    orientation='vertical')

    image_legend_label = ctx.get_legend_label(ds_id, var_name)
    if image_legend_label is not None:
        image_legend.set_label(image_legend_label)

    fig.patch.set_facecolor('white')
    fig.patch.set_alpha(0.0)
    fig.tight_layout()

    buffer = io.BytesIO()
    fig.savefig(buffer, format='png')

    return buffer.getvalue()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import inspect
import threading
from typing import Type, TypeVar, Mapping, Sequence, Any, Optional, Callable, \
    Tuple, Union

from xcube.util.assertions import assert_true

GET_EXTENSIONS_ATTR_NAME = 'get_extensions'

Base = TypeVar('Base')
Ext = TypeVar('Ext')

BaseClass = Type[Base]
ExtClass = Type[Ext]

ExtClassItem = Tuple[str, ExtClass]
ExtInstItem = Tuple[str, Ext]


def extend(
        base_class: BaseClass,
        name: str,
        doc: Optional[str] = None,
        class_handler: Union[None, str, Callable[[ExtClassItem], Any]] = None,
        inst_handler: Union[None, str, Callable[[ExtInstItem], Any]] = None,
        ext_args: Optional[Sequence[Any]] = None,
        ext_kwargs: Optional[Mapping[str, Any]] = None
) -> Callable[[ExtClass], ExtClass]:
    """
    A class decorator factory that adds an instance property named
    *property_name* to an existing class *base_class*. The value of the
    new property is an instance of the decorated class.

    This decorator can be used to add new functionality to existing
    classes without deriving from the existing class.

    The optional *class_handler* callback function can be used for
    registering the extensions in the base class and/or for
    validating the passed extension name and class.
    Likewise, the optional *inst_handler* callback function can be used for
    registering the extensions in the base class instances and/or for
    validating the passed extension name and instances.

    The decorated class is expected to have a constructor whose first
    positional argument is an instance of *base_class*.
    Other positional and keyword arguments may follow
    and may be passed to this decorator using the *ext_args*
    and *ext_kwargs* keyword arguments.

    :param base_class: The base class to be extended.
    :param name: The new property's name.
    :param doc: The new property's docstring, optional.
        If omitted, the docstring of the decorated class, if any, is used.
    :param class_handler: Optional class handler.
        If given, it must be a callable that is called
        with a tuple comprising extension name and class.
        If this is a string, it is the name of a static or class method
        of the *base_class* which is called with that tuple.
    :param inst_handler: Optional instance handler.
        If given, it must be a callable that is called
        with a tuple comprising extension name and instance.
        If this is a string, it is the name of an instance method
        of the *base_class* which is called with that tuple.
    :param ext_args: positional arguments passed
        to the decorated class' constructor
    :param ext_kwargs: keyword arguments passed
        to the decorated class' constructor
    :return: a class decorator function
    """
    assert_true(inspect.isclass(base_class),
                message=f'base_class must be a class type,'
                        f' but was {type(base_class).__name__}')
    assert_true(isinstance(name, str)
                and name.isidentifier(),
                message=f'name must be a valid identifier,'
                        f' but was {name!r}')
    assert_true(doc is None or isinstance(doc, str),
                message='doc must be a string')

    _name = '_' + name
    lock = threading.RLock()

    def call_handler(base, handler, ext):
        if isinstance(handler, str):
            getattr(base, handler)((name, ext))
        elif callable(handler):
            handler((name, ext))

    def add_class_property(ext_class: ExtClass) -> ExtClass:
        assert_true(inspect.isclass(ext_class),
                    message='the extend() decorator can'
                            ' be used with classes only')

        def _get_property(base: Base):
            ext = getattr(base, _name, None)
            if ext is None:
                with lock:
                    ext = getattr(base, _name, None)
                    if ext is None:
                        ext = ext_class(base,
                                        *(ext_args or ()),
                                        **(ext_kwargs or {}))
                        setattr(base, _name, ext)
                        call_handler(base, inst_handler, ext)
            return ext

        assert_true(not hasattr(base_class, name),
                    message=f'a property named {name} already'
                            f' exists in class {base_class.__name__}')

        setattr(base_class,
                name,
                property(_get_property, None, None,
                         ext_class.__doc__ if doc is None else doc))

        call_handler(base_class, class_handler, ext_class)

        return ext_class

    return add_class_property



# The MIT License (MIT)
# Copyright (c) 2021/2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Optional, Tuple, Dict, Any

import fsspec
import rasterio
import xarray as xr

from xcube.core.mldataset import LazyMultiLevelDataset
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.store import DataType
from xcube.core.store import MULTI_LEVEL_DATASET_TYPE
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from .dataset import DatasetGeoTiffFsDataAccessor


class GeoTIFFMultiLevelDataset(LazyMultiLevelDataset):
    """
    A multi-level dataset for GeoTIFF format

    @param fs: fsspec.AbstractFileSystem object.
    @param root: Optional root path identifier.
    @param data_id: dataset identifier.
    @param open_params: keyword arguments.
    """

    def __init__(self,
                 fs: fsspec.AbstractFileSystem,
                 root: Optional[str],
                 data_id: str,
                 **open_params: Dict[str, Any]):
        super().__init__(ds_id=data_id)
        self._fs = fs
        self._root = root
        self._path = data_id
        self._open_params = open_params
        self._file_url = None

    def _get_overview_count(self):
        with rasterio.open(self._file_url) as rio_dataset:
            overviews = rio_dataset.overviews(1)
        return overviews

    def _get_num_levels_lazily(self) -> int:
        self._file_url = self._get_file_url()
        if isinstance(self._fs, fsspec.AbstractFileSystem):
            with DatasetGeoTiffFsDataAccessor.create_env_session(self._fs):
                overviews = self._get_overview_count()
        else:
            assert_true(self._fs is None, message="invalid type for fs")
        return len(overviews) + 1

    def _get_dataset_lazily(self, index: int, parameters) \
            -> xr.Dataset:
        tile_size = self._open_params.get("tile_size", (512, 512))
        self._file_url = self._get_file_url()
        return DatasetGeoTiffFsDataAccessor.open_dataset(
            self._fs,
            self._file_url,
            tile_size,
            overview_level=index - 1 if index > 0 else None
        )

    def _get_file_url(self):
        if isinstance(self._fs.protocol, str):
            protocol = self._fs.protocol
        else:
            protocol = self._fs.protocol[0]
        return protocol + "://" + self._path


MULTI_LEVEL_GEOTIFF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        tile_size=JsonArraySchema(
            items=(
                JsonNumberSchema(minimum=256, default=512),
                JsonNumberSchema(minimum=256, default=512)
            ),
            default=[512, 512]
        ),
    ),
    additional_properties=False,
)


# noinspection PyAbstractClass
class MultiLevelDatasetGeoTiffFsDataAccessor(DatasetGeoTiffFsDataAccessor):
    """
    Opener/writer extension name: "mldataset:geotiff:<protocol>"
    """

    @classmethod
    def get_data_type(cls) -> DataType:
        return MULTI_LEVEL_DATASET_TYPE

    def get_open_data_params_schema(self,
                                    data_id: str = None) -> JsonObjectSchema:
        return MULTI_LEVEL_GEOTIFF_OPEN_DATA_PARAMS_SCHEMA

    def open_data(self, data_id: str, **open_params) -> MultiLevelDataset:
        assert_instance(data_id, str, name='data_id')
        fs, root, open_params = self.load_fs(open_params)
        return GeoTIFFMultiLevelDataset(fs, root, data_id, **open_params)



from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping

from xcube.webapi.compute.op.decorator import operation
from xcube.webapi.compute.op.decorator import op_param


@operation
@op_param("bbox",
          title="Bounding box",
          description="Bounding box using the dataset's CRS coordinates",
          # The rest of the schema is inferred from the function signature.
          )
def spatial_subset(dataset: xr.Dataset,
                   bbox: Tuple[float, float, float, float]) -> xr.Dataset:
    """Create a spatial subset from given dataset."""
    x1, y1, x2, y2 = bbox
    gm = GridMapping.from_dataset(dataset)
    x_name, y_name = gm.xy_dim_names
    return dataset.sel({
        x_name: slice(x1, x2),
        y_name: slice(y1, y2) if gm.is_j_axis_up else slice(y2, y1)
    })



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import traceback

from xcube.core.normalize import DatasetIsNotACubeError
from xcube.core.normalize import decode_cube
from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataStoreError
from xcube.core.store import DataStorePool
from xcube.core.store import get_data_store_instance
from xcube.core.store import new_data_opener
from xcube.util.assertions import assert_instance
from xcube.util.progress import observe_progress
from .transformer import TransformedCube
from ..config import CubeConfig
from ..config import InputConfig
from ..error import CubeGeneratorError

# Names of cube configuration parameters that
# are not shared with open parameters.
_STEADY_CUBE_CONFIG_NAMES = {
    'chunks',
    'tile_size'
}


class CubeOpener:

    def __init__(self,
                 cube_config: CubeConfig,
                 store_pool: DataStorePool = None):
        assert_instance(cube_config, CubeConfig, 'cube_config')
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, 'store_pool')
        self._cube_config = cube_config
        self._store_pool = store_pool

    def open_cube(self, input_config: InputConfig) -> TransformedCube:
        cube_config = self._cube_config
        cube_params = cube_config.to_dict()
        opener_id = input_config.opener_id
        store_params = input_config.store_params or {}
        open_params = input_config.open_params or {}

        with observe_progress('reading cube', 3) as observer:
            try:
                if input_config.store_id:
                    store_instance = get_data_store_instance(
                        input_config.store_id,
                        store_params=store_params,
                        store_pool=self._store_pool
                    )
                    store = store_instance.store
                    if opener_id is None:
                        opener_id = self._get_opener_id(
                            input_config, store
                        )
                    opener = store
                    open_params = dict(open_params)
                    open_params['opener_id'] = opener_id
                else:
                    opener = new_data_opener(opener_id)
                    open_params = dict(open_params)
                    open_params.update(store_params)

                open_params_schema = opener.get_open_data_params_schema(
                    input_config.data_id
                )

                dataset_open_params = {
                    k: v for k, v in cube_params.items()
                    if k in open_params_schema.properties
                }

                observer.worked(1)

                dataset = opener.open_data(input_config.data_id,
                                           **open_params,
                                           **dataset_open_params)
                observer.worked(1)

            except DataStoreError as dse:
                raise CubeGeneratorError(f'{dse}', status_code=400) from dse

            # Turn dataset into cube and grid_mapping
            try:
                cube, gm, _ = decode_cube(dataset,
                                          normalize=True)
            except DatasetIsNotACubeError as e:
                raise CubeGeneratorError(f'{e}') from e
            observer.worked(1)

        if dataset_open_params:
            drop_names = [k for k in dataset_open_params.keys()
                          if k not in _STEADY_CUBE_CONFIG_NAMES]
            cube_config = cube_config.drop_props(drop_names)

        return cube, gm, cube_config

    @classmethod
    def _get_opener_id(cls, input_config, store) -> str:
        opener_ids = None
        data_type_names = store.get_data_types_for_data(
            input_config.data_id
        )
        for data_type_name in data_type_names:
            if DATASET_TYPE.is_super_type_of(data_type_name):
                opener_ids = \
                    store.get_data_opener_ids(
                        data_id=input_config.data_id,
                        data_type=data_type_name
                    )
                break
        if not opener_ids:
            raise CubeGeneratorError(f'Data store {input_config.store_id!r}'
                                     f' does not support datasets',
                                     status_code=400)
        opener_id = opener_ids[0]
        return opener_id



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

import click

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

DEFAULT_OUTPUT_PATH = 'out.zarr'


@click.command(name='compute')
@click.argument('script')
@click.argument('cube', nargs=-1)
@click.option('--variables', '--vars', 'input_var_names', metavar='VARIABLES',
              help="Comma-separated list of variable names.")
@click.option('--params', '-p', 'input_params', metavar='PARAMS',
              help="Parameters passed as 'input_params' dict to compute() and init() functions in SCRIPT.")
@click.option('--output', '-o', 'output_path', metavar='OUTPUT', default=DEFAULT_OUTPUT_PATH,
              help=f'Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}')
@click.option('--format', '-f', 'output_format', metavar='FORMAT',
              default='zarr',
              type=click.Choice(['zarr', 'nc']),
              help="Output format.")
@click.option('--name', '-N', 'output_var_name', metavar='NAME',
              default='output',
              help="Output variable's name.")
@click.option('--dtype', '-D', 'output_var_dtype', metavar='DTYPE',
              default='float64',
              type=click.Choice(["uint8", "int8", "uint16", "int16", "uint32", "int32", "float32", "float64"]),
              help="Output variable's data type.")
def compute(script: str,
            cube: List[str],
            input_var_names: str,
            input_params: str,
            output_path: str,
            output_format: str,
            output_var_name: str,
            output_var_dtype: str):
    """
    Compute a cube from one or more other cubes.

    The command computes a cube variable from other cube variables in CUBEs
    using a user-provided Python function in SCRIPT.

    The SCRIPT must define a function named "compute":

    \b
        def compute(*input_vars: numpy.ndarray,
                    input_params: Mapping[str, Any] = None,
                    dim_coords: Mapping[str, np.ndarray] = None,
                    dim_ranges: Mapping[str, Tuple[int, int]] = None) \\
                    -> numpy.ndarray:
            # Compute new numpy array from inputs
            # output_array = ...
            return output_array

    where input_vars are numpy arrays (chunks) in the order given by VARIABLES or given by the variable names returned
    by an optional "initialize" function that my be defined in SCRIPT too, see below. input_params is a mapping of
    parameter names to values according to PARAMS or the ones returned by the aforesaid "initialize" function.
    dim_coords is a mapping from dimension name to coordinate labels for the current chunk to be computed.
    dim_ranges is a mapping from dimension name to index ranges into coordinate arrays of the cube.

    The SCRIPT may define a function named "initialize":

    \b
        def initialize(input_cubes: Sequence[xr.Dataset],
                       input_var_names: Sequence[str],
                       input_params: Mapping[str, Any]) \\
                       -> Tuple[Sequence[str], Mapping[str, Any]]:
            # Compute new variable names and/or new parameters
            # new_input_var_names = ...
            # new_input_params = ...
            return new_input_var_names, new_input_params

    where input_cubes are the respective CUBEs, input_var_names the respective VARIABLES, and input_params
    are the respective PARAMS. The "initialize" function can be used to validate the data cubes, extract
    the desired variables in desired order and to provide some extra processing parameters passed to the
    "compute" function.

    Note that if no input variable names are specified, no variables are passed to the "compute" function.

    The SCRIPT may also define a function named "finalize":

    \b
        def finalize(output_cube: xr.Dataset,
                     input_params: Mapping[str, Any]) \\
                     -> Optional[xr.Dataset]:
            # Optionally modify output_cube and return it or return None
            return output_cube

    If defined, the "finalize" function will be called before the command writes the
    new cube and then exists. The functions may perform a cleaning up or perform side effects such
    as write the cube to some sink. If the functions returns None, the CLI will *not* write
    any cube data.

    """
    from xcube.cli.common import parse_cli_kwargs
    from xcube.core.compute import compute_cube
    from xcube.core.dsio import open_cube
    from xcube.core.dsio import guess_dataset_format, find_dataset_io

    input_paths = cube

    compute_function_name = "compute"
    initialize_function_name = "initialize"
    finalize_function_name = "finalize"

    with open(script, "r") as fp:
        code = fp.read()

    locals_dict = dict()
    exec(code, globals(), locals_dict)

    input_var_names = list(map(lambda s: s.strip(), input_var_names.split(","))) if input_var_names else None

    compute_function = _get_function(locals_dict, compute_function_name, script, force=True)
    initialize_function = _get_function(locals_dict, initialize_function_name, script, force=False)
    finalize_function = _get_function(locals_dict, finalize_function_name, script, force=False)

    input_params = parse_cli_kwargs(input_params, "PARAMS")

    input_cubes = []
    for input_path in input_paths:
        input_cubes.append(open_cube(input_path=input_path))

    if initialize_function:
        input_var_names, input_params = initialize_function(input_cubes, input_var_names, input_params)

    output_cube = compute_cube(compute_function,
                               *input_cubes,
                               input_var_names=input_var_names,
                               input_params=input_params,
                               output_var_name=output_var_name,
                               output_var_dtype=output_var_dtype)

    if finalize_function:
        output_cube = finalize_function(output_cube)

    if output_cube is not None:
        output_format = output_format or guess_dataset_format(output_path)
        dataset_io = find_dataset_io(output_format, {"w"})
        dataset_io.write(output_cube, output_path)


def _get_function(object, function_name, source, force=False):
    function = object.get(function_name)
    if function is None:
        if force:
            raise click.ClickException(f"missing function {function_name!r} in {source}")
        else:
            return None
    if not callable(function):
        raise click.ClickException(f"{function_name!r} in {source} is not a callable")
    return function



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import datetime
import fnmatch
import io
import json
import os
import os.path
import string
from typing import Any, Dict, Optional, Iterable, Tuple, List, Type

import fsspec
import yaml

from xcube.constants import LOG

PRIMITIVE_TYPES = (int, float, str, type(None))

NameAnyDict = Dict[str, Any]
NameDictPairList = List[Tuple[str, Optional[NameAnyDict]]]


def to_resolved_name_dict_pairs(name_dict_pairs: NameDictPairList,
                                container,
                                keep=False) -> NameDictPairList:
    resolved_pairs = []
    for name, value in name_dict_pairs:
        if '*' in name or '?' in name or '[' in name:
            for resolved_name in container:
                if fnmatch.fnmatch(resolved_name, name):
                    resolved_pairs.append((resolved_name, value))
        elif name in container or keep:
            resolved_pairs.append((name, value))
    return resolved_pairs


def to_name_dict_pairs(iterable: Iterable[Any], default_key=None) \
        -> NameDictPairList:
    return [to_name_dict_pair(item, parent=iterable, default_key=default_key)
            for item in iterable]


def to_name_dict_pair(name: Any, parent: Any = None, default_key=None):
    value = None

    if isinstance(name, str):
        try:
            # is key of parent?
            value = parent[name]
        except (KeyError, TypeError, ValueError):
            pass
    else:
        # key = (key, value)?
        try:
            name, value = name
        except (TypeError, ValueError):
            # key = {key: value}?
            try:
                name, value = dict(name).popitem()
            except (TypeError, ValueError, AttributeError, KeyError):
                pass

    if not isinstance(name, str):
        raise ValueError(f'name must be a string')

    if value is None:
        return name, None

    try:
        # noinspection PyUnresolvedReferences
        value.items()
    except AttributeError as e:
        if default_key:
            value = {default_key: value}
        else:
            raise ValueError(f'value of {name!r} must be a dictionary') from e

    return name, value


def flatten_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    result = dict()
    value = _flatten_dict_value(d, result, None, False)
    if value is not result:
        raise ValueError('input must be a mapping object')
    # noinspection PyTypeChecker
    return result


def _flatten_dict_value(value: Any,
                        result: Dict[str, Any],
                        parent_name: Optional[str],
                        concat: bool) -> Any:
    if isinstance(value, datetime.datetime):
        return datetime.datetime.isoformat(value)
    elif isinstance(value, datetime.date):
        return datetime.date.isoformat(value)
    elif isinstance(value, PRIMITIVE_TYPES):
        return value

    try:
        items = value.items()
    except AttributeError:
        items = None

    if items:
        for k, v in items:
            if not isinstance(k, str):
                raise ValueError('all keys must be strings')
            v = _flatten_dict_value(v, result, k, False)
            if v is not result:
                name = k if parent_name is None else parent_name + '_' + k
                if concat and name in result:
                    result[name] = f'{result[name]}, {v}'
                else:
                    result[name] = v

    for e in value:
        _flatten_dict_value(e, result, parent_name, True)

    return result


def merge_config(first_dict: Dict, *more_dicts):
    if not more_dicts:
        output_dict = first_dict
    else:
        output_dict = dict(first_dict)
        for d in more_dicts:
            for k, v in d.items():
                if k in output_dict \
                        and isinstance(output_dict[k], dict) \
                        and isinstance(v, dict):
                    v = merge_config(output_dict[k], v)
                output_dict[k] = v
    return output_dict


def load_configs(
        *config_paths: str,
        exception_type: Type[Exception] = ValueError
) -> Dict[str, Any]:
    config_dicts = []
    for config_path in config_paths:
        config_dict = load_json_or_yaml_config(
            config_path,
            exception_type=exception_type
        )
        config_dicts.append(config_dict)
    config = merge_config(*config_dicts)
    return config


def load_json_or_yaml_config(
        config_path: str,
        exception_type: Type[Exception] = ValueError
) -> Dict[str, Any]:
    try:
        config_dict = _load_json_or_yaml_config(config_path)
        LOG.info(f'Configuration loaded: {config_path}')
    except FileNotFoundError as e:
        raise exception_type(
            f'Cannot find configuration {config_path!r}'
        ) from e
    except yaml.YAMLError as e:
        raise exception_type(
            f'YAML in {config_path!r} is invalid: {e}'
        ) from e
    except OSError as e:
        raise exception_type(
            f'Cannot load configuration from'
            f' {config_path!r}: {e}'
        ) from e
    if config_dict is None:
        return {}
    if not isinstance(config_dict, dict):
        raise exception_type(
            f'Invalid configuration format in'
            f' {config_path!r}: dictionary expected'
        )
    return config_dict


def _load_json_or_yaml_config(config_file: str) -> Any:
    with fsspec.open(config_file, mode='r') as fp:
        file_content = fp.read()
    template = string.Template(file_content)
    file_content = template.safe_substitute(os.environ)
    with io.StringIO(file_content) as fp:
        if config_file.endswith('.json'):
            return json.load(fp)
        else:
            return yaml.safe_load(fp)



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Sequence

from xcube.util.config import flatten_dict, load_configs, to_name_dict_pairs


def get_config_dict(config_files: Sequence[str] = None,
                    input_paths: Sequence[str] = None,
                    input_processor_name: str = None,
                    output_path: str = None,
                    output_writer_name: str = None,
                    output_size: str = None,
                    output_region: str = None,
                    output_variables: str = None,
                    output_resampling: str = None,
                    append_mode: bool = True,
                    profile_mode: bool = False,
                    no_sort_mode: bool = False):
    """
    Get a configuration dictionary from given (command-line) arguments.

    :return: Configuration dictionary
    :raise OSError, ValueError
    """

    config = load_configs(*config_files) if config_files else {}

    # preserve backward compatibility for old names
    if 'input_processor' in config:
        config['input_processor_name'] = config.pop('input_processor')
    if 'output_writer' in config:
        config['output_writer_name'] = config.pop('output_writer')

    # Overwrite current configuration by cli arguments
    if input_paths is not None and 'input_paths' not in config:
        if len(input_paths) == 1 and input_paths[0].endswith(".txt"):
            with open(input_paths[0]) as input_txt:
                input_paths = input_txt.readlines()
            config['input_paths'] = [x.strip() for x in input_paths]
        else:
            config['input_paths'] = input_paths

    if input_processor_name is not None and 'input_processor_name' not in config:
        config['input_processor_name'] = input_processor_name

    if output_path is not None and 'output_path' not in config:
        config['output_path'] = output_path

    if output_writer_name is not None and 'output_writer_name' not in config:
        config['output_writer_name'] = output_writer_name

    if output_resampling is not None and 'output_resampling' not in config:
        config['output_resampling'] = output_resampling

    if output_size is not None:
        try:
            output_size = list(map(lambda c: int(c), output_size.split(',')))
        except ValueError:
            output_size = None
        if output_size is None or len(output_size) != 2:
            raise ValueError(f'output_size must have the form <width>,<height>,'
                             f' where both values must be positive integer numbers')
        config['output_size'] = output_size

    if output_region is not None:
        try:
            output_region = list(map(lambda c: float(c), output_region.split(',')))
        except ValueError:
            output_region = None
        if output_region is None or len(output_region) != 4:
            raise ValueError(f'output_region must have the form <lon_min>,<lat_min>,<lon_max>,<lat_max>,'
                             f' where all four numbers must be floating point numbers in degrees')
        config['output_region'] = output_region

    if output_variables is not None:
        output_variables = list(map(lambda c: c.strip(), output_variables.split(',')))
        if output_variables == [''] or any([var_name == '' for var_name in output_variables]):
            raise ValueError('output_variables must be a list of existing variable names')
        config['output_variables'] = output_variables

    if profile_mode is not None and config.get('profile_mode') is None:
        config['profile_mode'] = profile_mode

    if append_mode is not None and config.get('append_mode') is None:
        config['append_mode'] = append_mode

    if no_sort_mode is not None and config.get('no_sort_mode') is None:
        config['no_sort_mode'] = no_sort_mode

    processed_variables = config.get('processed_variables')
    if processed_variables:
        config['processed_variables'] = to_name_dict_pairs(processed_variables)

    output_variables = config.get('output_variables')
    if output_variables:
        config['output_variables'] = to_name_dict_pairs(output_variables, default_key='name')

    output_metadata = config.get('output_metadata')
    if output_metadata:
        config['output_metadata'] = flatten_dict(output_metadata)

    return config



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import collections.abc
from logging import Logger
from typing import Iterator, Iterable, KeysView, Optional

import zarr.storage

from xcube.constants import LOG
from xcube.util.assertions import assert_instance
from xcube.util.perf import measure_time_cm


class LoggingZarrStore(zarr.storage.BaseStore):
    """
    A Zarr Store that logs all method calls on another store *other*
    including execution time.
    """

    def __init__(self,
                 other: collections.abc.MutableMapping,
                 logger: Logger = LOG,
                 name: Optional[str] = None):
        assert_instance(other, collections.abc.MutableMapping)
        self._other = other
        self._measure_time = measure_time_cm(logger=logger)
        self._name = name or 'chunk_store'
        if hasattr(other, 'listdir'):
            setattr(self, 'listdir', self.__listdir)
        if hasattr(other, 'getsize'):
            setattr(self, 'getsize', self.__getsize)

    def __listdir(self, key: str) -> Iterable[str]:
        with self._measure_time(f'{self._name}.listdir({key!r})'):
            # noinspection PyUnresolvedReferences
            return self._other.listdir(key)

    def __getsize(self, key: str) -> int:
        with self._measure_time(f'{self._name}.getsize({key!r})'):
            # noinspection PyUnresolvedReferences
            return self._other.getsize(key)

    def keys(self) -> KeysView[str]:
        with self._measure_time(f'{self._name}.keys()'):
            # noinspection PyTypeChecker
            return self._other.keys()

    def __iter__(self) -> Iterator[str]:
        with self._measure_time(f'{self._name}.__iter__()'):
            return self._other.__iter__()

    def __len__(self) -> int:
        with self._measure_time(f'{self._name}.__len__()'):
            return self._other.__len__()

    def __contains__(self, key) -> bool:
        with self._measure_time(f'{self._name}.__contains__({key!r})'):
            return self._other.__contains__(key)

    def __getitem__(self, key: str) -> bytes:
        with self._measure_time(f'{self._name}.__getitem__({key!r})'):
            return self._other.__getitem__(key)

    def __setitem__(self, key: str, value: bytes) -> None:
        with self._measure_time(f'{self._name}.__setitem__({key!r}, <value>)'):
            return self._other.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        with self._measure_time(f'{self._name}.__delitem__({key!r})'):
            return self._other.__delitem__(key)



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Any, Union, Tuple, Type, Container

_DEFAULT_NAME = 'value'


def assert_not_none(value: Any, name: str = None,
                    exception_type: Type[Exception] = ValueError):
    """
    Assert *value* is not None.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param name: Name of a variable that holds *value*.
    :param exception_type: The exception type. Default is ```ValueError```.
    """
    if value is None:
        raise exception_type(f'{name or _DEFAULT_NAME} must not be None')


def assert_given(value: Any,
                 name: str = None,
                 exception_type: Type[Exception] = ValueError):
    """
    Assert *value* is not False when converted into a Boolean value.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param name: Name of a variable that holds *value*.
    :param exception_type: The exception type. Default is ```ValueError```.
    """
    if not value:
        raise exception_type(f'{name or _DEFAULT_NAME} must be given')


def assert_instance(value: Any,
                    dtype: Union[Type, Tuple[Type, ...]],
                    name: str = None,
                    exception_type: Type[Exception] = TypeError):
    """
    Assert *value* is an instance of data type *dtype*.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param dtype: A type or tuple of types.
    :param name: Name of a variable that holds *value*.
    :param exception_type: The exception type. Default is ```TypeError```.
    """
    if not isinstance(value, dtype):
        raise exception_type(f'{name or _DEFAULT_NAME} '
                             f'must be an instance of '
                             f'{dtype}, was {type(value)}')


def assert_subclass(value: Any,
                    cls: Union[Type, Tuple[Type, ...]],
                    name: str = None,
                    exception_type: Type[Exception] = TypeError):
    """
    Assert *value* is a subclass of class *cls*.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param cls: A class or tuple of classes.
    :param name: Name of a variable that holds *value*.
    :param exception_type: The exception type. Default is ```TypeError```.
    """
    if not issubclass(value, cls):
        raise exception_type(f'{name or _DEFAULT_NAME} '
                             f'must be a subclass of '
                             f'{cls}, was {value}')


def assert_in(value: Any,
              container: Container,
              name: str = None,
              exception_type: Type[Exception] = ValueError):
    """
    Assert *value* is a member of *container*.
    Otherwise, raise *exception_type*.

    :param value: The value to test for membership.
    :param container: The container.
    :param name: Name of a variable that holds *value*.
    :param exception_type: The exception type. Default is ```ValueError```.
    """
    if value not in container:
        raise exception_type(f'{name or _DEFAULT_NAME} '
                             f'must be one of {container}')


def assert_true(value: Any,
                message: str,
                exception_type: Type[Exception] = ValueError):
    """
    Assert *value* is true after conversion into a Boolean value.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param message: The error message used if the assertion fails.
    :param exception_type: The exception type. Default is ```ValueError```.
    """
    if not value:
        raise exception_type(message)


def assert_false(value: Any,
                 message: str,
                 exception_type: Type[Exception] = ValueError):
    """
    Assert *value* is false after conversion into a Boolean value.
    Otherwise, raise *exception_type*.

    :param value: The value to test.
    :param message: The error message used if the assertion fails.
    :param exception_type: The exception type. Default is ```ValueError```.
    """
    if value:
        raise exception_type(message)


def assert_condition(condition: Any,
                     message: str,
                     exception_type: Type[Exception] = ValueError):
    """Deprecated. Use assert_true()"""
    warnings.warn('assert_condition() has been deprecated. '
                  'Use assert_true() or assert_false() instead.',
                  DeprecationWarning)
    if not condition:
        raise exception_type(message)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC, abstractmethod
from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.progress import observe_progress
from .helpers import is_empty_cube
from .helpers import strip_cube
from ..config import CubeConfig

TransformedCube = Tuple[xr.Dataset, GridMapping, CubeConfig]


class CubeTransformer(ABC):
    @abstractmethod
    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        """
        Transform given *cube*, grid mapping *gm*, and cube
        configuration *cube_config* into a potentially new cube,
        a new grid mapping, and less restrictive cube configuration
        *cube_config*.

        The latter is achieved by returning a derived *cube_config* where
        all properties that have been "consumed" by this transformer
        are removed. See :meth:`CubeConfig.drop_property`.
        """


class CubeIdentity(CubeTransformer):
    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        """
        Return *cube*, grid mapping *gm*, and parameters without change.
        """
        return cube, gm, cube_config


def transform_cube(t_cube: TransformedCube,
                   transformer: CubeTransformer,
                   label: str = '') -> TransformedCube:
    empty_cube = is_empty_cube(t_cube[0])
    identity = isinstance(transformer, CubeIdentity)
    if not label:
        label = f'{type(transformer).__name__}'
    if identity:
        label += ' (step not applicable)'
    elif empty_cube:
        label += ' (step not applicable, empty cube)'

    with observe_progress(label, 1) as progress:
        if not (identity or empty_cube):
            t_cube = transformer.transform_cube(*t_cube)
            t_cube = strip_cube(t_cube[0]), t_cube[1], t_cube[2]
        progress.worked(1)

    return t_cube



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import tempfile
from collections.abc import MutableMapping
from typing import Dict, Union, Tuple

import numpy as np
import xarray as xr
import zarr

from xcube.core.chunk import chunk_dataset
from xcube.core.unchunk import unchunk_dataset

DEFAULT_TIME_EPS = np.array(1000 * 1000, dtype='timedelta64[ns]')


def find_time_slice(store: Union[str, MutableMapping],
                    time_stamp: Union[np.datetime64, np.ndarray],
                    time_eps: np.timedelta64 = DEFAULT_TIME_EPS) -> Tuple[int, str]:
    """
    Find time index and update mode for *time_stamp* in ZARR dataset given by *store*.

    :param store: A zarr store.
    :param time_stamp: Time stamp to find index for.
    :param time_eps: Time epsilon for equality comparison, defaults to 1 millisecond.
    :return: A tuple (time_index, 'insert') or (time_index, 'replace') if an index was found,
        (-1, 'create') or (-1, 'append') otherwise.
    """
    try:
        cube = xr.open_zarr(store)
    except (FileNotFoundError, ValueError):
        # FileNotFoundError is raised as by Zarr since 2.13,
        # before GroupNotFoundError (extends ValueError) was raised.
        # Keep ValueError for backward compatibility.
        try:
            cube = xr.open_dataset(store)
        except (FileNotFoundError, ValueError):
            # If the zarr directory does not exist, open_dataset raises a
            # FileNotFoundError (with xarray <= 0.17.0) or a ValueError
            # (with xarray 0.18.0).
            return -1, 'create'

    # TODO (forman): optimise following naive search by bi-sectioning or so
    for i in range(cube.time.size):
        time = cube.time[i]
        if abs(time_stamp - time) < time_eps:
            return i, 'replace'
        if time_stamp < time:
            return i, 'insert'

    return -1, 'append'


def append_time_slice(store: Union[str, MutableMapping],
                      time_slice: xr.Dataset,
                      chunk_sizes: Dict[str, int] = None):
    """
    Append time slice to existing zarr dataset.

    :param store: A zarr store.
    :param time_slice: Time slice to insert
    :param chunk_sizes: desired chunk sizes
    """
    if chunk_sizes:
        time_slice = chunk_dataset(time_slice, chunk_sizes, format_name='zarr')

    # Unfortunately time_slice.to_zarr(store, mode='a', append_dim='time') will replace global attributes of store
    # with attributes of time_slice (xarray bug?), which are usually empty in our case.
    # Hence, we must save our old attributes in a copy of time_slice.
    ds = zarr.open_group(store, mode='r')
    time_slice = time_slice.copy()
    time_slice.attrs.update(ds.attrs)
    if 'coordinates' in time_slice.attrs:
        # Remove 'coordinates', otherwise we get
        # ValueError: cannot serialize coordinates because the global attribute 'coordinates' already exists
        # from next time_slice.to_zarr(...) call.
        time_slice.attrs.pop('coordinates')

    time_slice.to_zarr(store, mode='a', append_dim='time', consolidated=True)

    unchunk_dataset(store, coords_only=True)


def insert_time_slice(store: Union[str, MutableMapping],
                      insert_index: int,
                      time_slice: xr.Dataset,
                      chunk_sizes: Dict[str, int] = None):
    """
    Insert time slice into existing zarr dataset.

    :param store: A zarr store.
    :param insert_index: Time index
    :param time_slice: Time slice to insert
    :param chunk_sizes: desired chunk sizes
    """
    update_time_slice(store, insert_index, time_slice, 'insert', chunk_sizes=chunk_sizes)


def replace_time_slice(store: Union[str, MutableMapping],
                       insert_index: int,
                       time_slice: xr.Dataset,
                       chunk_sizes: Dict[str, int] = None):
    """
    Replace time slice in existing zarr dataset.

    :param store: A zarr store.
    :param insert_index: Time index
    :param time_slice: Time slice to insert
    :param chunk_sizes: desired chunk sizes
    """
    update_time_slice(store, insert_index, time_slice, 'replace', chunk_sizes=chunk_sizes)


def update_time_slice(store: Union[str, MutableMapping],
                      insert_index: int,
                      time_slice: xr.Dataset,
                      mode: str,
                      chunk_sizes: Dict[str, int] = None):
    """
    Update existing zarr dataset by new time slice.

    :param store: A zarr store.
    :param insert_index: Time index
    :param time_slice: Time slice to insert
    :param mode: Update mode, 'insert' or 'replace'
    :param chunk_sizes: desired chunk sizes
    """

    if mode not in ('insert', 'replace'):
        raise ValueError(f'illegal mode value: {mode!r}')

    insert_mode = mode == 'insert'

    time_var_names = []
    encoding = {}
    with xr.open_zarr(store) as cube:
        for var_name in cube.variables:
            var = cube[var_name]
            if var.ndim >= 1 and 'time' in var.dims:
                if var.dims[0] != 'time':
                    raise ValueError(f"dimension 'time' of variable {var_name!r} must be first dimension")
                time_var_names.append(var_name)
                enc = dict(cube[var_name].encoding)
                # xarray 0.17+ supports engine preferred chunks if exposed by the backend
                # zarr does that, but when we use the new 'preferred_chunks' when writing to zarr
                # it raises and says, 'preferred_chunks' is an unsupported encoding
                if 'preferred_chunks' in enc:
                    del enc['preferred_chunks']
                encoding[var_name] = enc

    if chunk_sizes:
        time_slice = chunk_dataset(time_slice, chunk_sizes, format_name='zarr')
    temp_dir = tempfile.TemporaryDirectory(prefix='xcube-time-slice-', suffix='.zarr')
    time_slice.to_zarr(temp_dir.name, encoding=encoding)
    slice_root_group = zarr.open(temp_dir.name, mode='r')
    slice_arrays = dict(slice_root_group.arrays())

    cube_root_group = zarr.open(store, mode='r+')
    for var_name, var_array in cube_root_group.arrays():
        if var_name in time_var_names:
            slice_array = slice_arrays[var_name]
            if insert_mode:
                # Add one empty time step
                empty = zarr.creation.empty(slice_array.shape, dtype=var_array.dtype)
                var_array.append(empty, axis=0)
                # Shift contents
                var_array[insert_index + 1:, ...] = var_array[insert_index:-1, ...]
            # Replace slice
            var_array[insert_index, ...] = slice_array[0]

    unchunk_dataset(store, coords_only=True)



#  The MIT License (MIT)
#  Copyright (c) 2022 by the xcube development team and contributors
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import collections.abc
import fnmatch
from typing import Dict, Tuple, Hashable, Optional, Mapping, Union

import numpy as np
import xarray as xr
from deprecated import deprecated

from xcube.util.assertions import assert_instance, assert_in, assert_true

AGG_METHODS = 'auto', 'first', 'min', 'max', 'mean', 'median'
DEFAULT_INT_AGG_METHOD = 'first'
DEFAULT_FLOAT_AGG_METHOD = 'mean'

AggMethod = Union[None, str]
AggMethods = Union[AggMethod, Mapping[str, AggMethod]]


def subsample_dataset(
        dataset: xr.Dataset,
        step: int,
        xy_dim_names: Optional[Tuple[str, str]] = None,
        agg_methods: Optional[AggMethods] = None
) -> xr.Dataset:
    """
    Subsample *dataset* with given integer subsampling *step*.
    Only data variables with spatial dimensions given by
    *xy_dim_names* are subsampled.

    :param dataset: the dataset providing the variables
    :param step: the integer subsampling step size in pixels in
        the x and y directions.
        For aggregation methods other than "first" it defines the
        window size for the aggregation.
    :param xy_dim_names: the spatial dimension names
    :param agg_methods: Optional aggregation methods.
        May be given as string or as mapping from variable name pattern
        to aggregation method. Valid aggregation methods are
        "auto", "first", "min", "max", "mean", "median".
        If "auto", the default, "first" is used for integer variables
        and "mean" for floating point variables.
    :return: The subsampled dataset or a tuple comprising the
        subsampled dataset and the effective aggregation methods.
    """
    assert_instance(dataset, xr.Dataset, name='dataset')
    assert_instance(step, int, name='step')
    assert_valid_agg_methods(agg_methods)

    x_name, y_name = xy_dim_names or ('y', 'x')

    agg_methods = get_dataset_agg_methods(dataset,
                                          xy_dim_names=xy_dim_names,
                                          agg_methods=agg_methods)

    new_data_vars = dict()
    new_coords = None  # used to collect coordinates from coarsen
    for var_name, var in dataset.data_vars.items():
        if var_name in agg_methods:
            agg_method = agg_methods[var_name]
            if agg_method == 'first':
                slices = get_variable_subsampling_slices(
                    var, step, xy_dim_names=xy_dim_names
                )
                assert slices is not None
                new_var = var[slices]
            else:
                dim = dict()
                if x_name in var.dims:
                    dim[x_name] = step
                if y_name in var.dims:
                    dim[y_name] = step
                var_coarsen = var.coarsen(dim=dim,
                                          boundary='pad',
                                          coord_func='min')
                new_var: xr.DataArray = getattr(var_coarsen, agg_method)()
                if new_var.dtype != var.dtype:
                    # We don't want, e.g. "mean", to turn data
                    # from dtype unit16 into float64
                    new_var = new_var.astype(var.dtype)
                new_var.attrs.update(var.attrs)
                new_var.encoding.update(var.encoding)
                # coarsen() recomputes spatial coordinates.
                # Collect them, so we can later apply them to the
                # variables that are subsampled by "first"
                # (= slice selection).
                if new_coords is None:
                    new_coords = dict(new_var.coords)
                else:
                    new_coords.update(new_var.coords)
        else:
            new_var = var

        new_data_vars[var_name] = new_var

    if not new_data_vars:
        return dataset

    if new_coords:
        # Make sure all variables use the same modified
        # spatial coordinates from coarsen
        new_data_vars = {
            k: v.assign_coords({
                d: new_coords[d]
                for d in v.dims if d in new_coords
            })
            for k, v in new_data_vars.items()
        }

    return xr.Dataset(data_vars=new_data_vars,
                      attrs=dataset.attrs)


def get_dataset_agg_methods(dataset: xr.Dataset,
                            xy_dim_names: Optional[Tuple[str, str]] = None,
                            agg_methods: Optional[AggMethods] = None):
    assert_instance(dataset, xr.Dataset, name='dataset')
    assert_valid_agg_methods(agg_methods)

    x_name, y_name = xy_dim_names or ('y', 'x')

    dataset_agg_methods = dict()

    for var_name, var in dataset.data_vars.items():
        if x_name in var.dims or y_name in var.dims:
            agg_method = find_agg_method(agg_methods, var_name, var.dtype)
            dataset_agg_methods[str(var_name)] = agg_method

    return dataset_agg_methods


def assert_valid_agg_methods(agg_methods: Optional[AggMethods]):
    """Assert that the given *agg_methods* are valid."""
    assert_instance(agg_methods,
                    (type(None), str, collections.abc.Mapping),
                    name='agg_methods')
    if isinstance(agg_methods, str):
        assert_in(
            agg_methods, AGG_METHODS,
            name='agg_methods'
        )
    elif agg_methods is not None:
        enum = (None, *AGG_METHODS)
        for k, v in agg_methods.items():
            assert_true(
                isinstance(k, str),
                message='keys in agg_methods must be strings'
            )
            assert_true(
                v in enum,
                message=f'values in agg_methods must be one of {enum}'
            )


def find_agg_method(agg_methods: AggMethods,
                    var_name: Hashable,
                    var_dtype: np.dtype) -> str:
    """
    Find aggregation method in *agg_methods*
    for given *var_name* and *var_dtype*.
    """
    assert_valid_agg_methods(agg_methods)
    if isinstance(agg_methods, str) and agg_methods != 'auto':
        return agg_methods
    if isinstance(agg_methods, collections.abc.Mapping):
        for var_name_pat, agg_method in agg_methods.items():
            if var_name == var_name_pat or fnmatch.fnmatch(str(var_name),
                                                           var_name_pat):
                if agg_method in (None, 'auto'):
                    break
                return agg_method
    # here: agg_method is either None or 'auto'
    if np.issubdtype(var_dtype, np.integer):
        return 'first'
    else:
        return 'mean'


_FULL_SLICE = slice(None, None, None)


def get_variable_subsampling_slices(
        variable: xr.DataArray,
        step: int,
        xy_dim_names: Optional[Tuple[str, str]] = None
) -> Optional[Tuple[slice, ...]]:
    """
    Compute subsampling slices for *variable*.
    Return None, if *variable* does not contain spatial
    dimensions.

    :param variable: the dataset providing the variables
    :param step: the integer subsampling step
    :param xy_dim_names: the spatial dimension names
    """
    assert_instance(variable, xr.DataArray, name='variable')
    assert_instance(step, int, name='step')
    x_dim_name, y_dim_name = xy_dim_names or ('x', 'y')

    var_index = None
    for index, dim_name in enumerate(variable.dims):
        if dim_name == x_dim_name or dim_name == y_dim_name:
            if var_index is None:
                var_index = index * [_FULL_SLICE]
            var_index.append(slice(None, None, step))
        elif var_index is not None:
            var_index.append(_FULL_SLICE)
    return tuple(var_index) if var_index is not None else None



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from datetime import datetime
from typing import Optional, Sequence, Union, Tuple

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from jdcal import jd2gcal

from xcube.core.gridmapping import GridMapping
from xcube.core.timecoord import get_timestamp_from_string
from xcube.core.timecoord import get_timestamps_from_string
from xcube.core.verify import assert_cube

DatetimeTypes = np.datetime64, cftime.datetime, datetime
Datetime = Union[np.datetime64, cftime.datetime, datetime]


class DatasetIsNotACubeError(BaseException):
    """
    Raised, if at least a subset of a dataset's variables
    have data cube dimensions ('time' , [...], y_dim_name, x_dim_name),
    where y_dim_name and x_dim_name are determined by a dataset's
    :class:GridMapping.
    """
    pass


def cubify_dataset(ds: xr.Dataset) -> xr.Dataset:
    """
    Normalize the geo- and time-coding upon opening the given
    dataset w.r.t. a common (CF-compatible) convention.

    Will throw a value error if the dataset could not not be
    converted to a cube.
    """
    ds = normalize_dataset(ds)
    return assert_cube(ds)


def normalize_dataset(ds: xr.Dataset,
                      reverse_decreasing_lat: bool = False) -> xr.Dataset:
    """
    Normalize the geo- and time-coding upon opening the given
    dataset w.r.t. a common (CF-compatible) convention.

    That is,
    * variables named "latitude" will be renamed to "lat";
    * variables named "longitude" or "long" will be renamed to "lon";

    Then, for equi-rectangular grids,
    * Remove 2D "lat" and "lon" variables;
    * Two new 1D coordinate variables "lat" and "lon" will be generated
    from original 2D forms.

    Then, if no coordinate variable time is present but the CF attributes
    "time_coverage_start"
    and optionally "time_coverage_end" are given, a scalar time dimension
    and coordinate variable
    will be generated.

    Finally, it will be ensured that a "time" coordinate variable will
    be of type *datetime*.

    :param ds: The dataset to normalize.
    :param reverse_decreasing_lat: Whether decreasing latitude values
        shall be normalized so they are increasing
    :return: The normalized dataset, or the original dataset, if it is
        already "normal".
    """
    ds = _normalize_zonal_lat_lon(ds)
    ds = normalize_coord_vars(ds)
    ds = _normalize_lat_lon(ds)
    ds = _normalize_lat_lon_2d(ds)
    ds = _normalize_dim_order(ds)
    ds = _normalize_lon_360(ds)
    if reverse_decreasing_lat:
        ds = _reverse_decreasing_lat(ds)
    ds = normalize_missing_time(ds)
    ds = _normalize_jd2datetime(ds)
    return ds


def encode_cube(cube: xr.Dataset,
                grid_mapping: Optional[GridMapping] = None,
                non_cube_subset: Optional[xr.Dataset] = None) \
        -> xr.Dataset:
    """
    Encode a *cube* with its *grid_mapping*, and additional variables in
    *non_cube_subset* into a new dataset.

    This is the inverse of the operation :func:decode_cube:

        cube, gm, non_cube = decode_cube(dataset)
        dataset = encode_cube(cube, gm, non_cube)

    The returned data cube comprises all variables in *cube*,
    whose dimensions should be ("time" , [...], y_dim_name, x_dim_name),
    and where y_dim_name, x_dim_name are defined by *grid_mapping*, if
    given.
    If *grid_mapping* is not geographic, a new variable "crs" will
    be added that holds CF-compliant attributes which encode the
    cube's spatial CRS. *non_cube_subset*, if given may be used
    to add non-cube variables the to resulting dataset.

    :param cube: data cube dataset, whose dimensions should
        be ("time" , [...], y_dim_name, x_dim_name)
    :param grid_mapping: Optional grid mapping for *cube*.
    :param non_cube_subset: An optional dataset providing
        non-cube data variables.
    :return:
    """
    if non_cube_subset is not None:
        dataset = cube.assign(**non_cube_subset.data_vars)
    else:
        dataset = cube

    if grid_mapping is None:
        return dataset

    if grid_mapping.crs.is_geographic \
            and grid_mapping.is_regular \
            and grid_mapping.xy_dim_names == ('lon', 'lat') \
            and grid_mapping.xy_var_names == ('lon', 'lat'):
        # No need to add CRS variable
        return dataset

    return dataset.assign(
        crs=xr.DataArray(0, attrs=grid_mapping.crs.to_cf())
    )


def decode_cube(dataset: xr.Dataset,
                normalize: bool = False,
                force_copy: bool = False,
                force_non_empty: bool = False,
                force_geographic: bool = False) \
        -> Tuple[xr.Dataset, GridMapping, xr.Dataset]:
    """
    Decode a *dataset* into a cube variable subset, a grid mapping, and
    the non-cube variables of *dataset*.

    This is the inverse of the operation :func:encode_cube:

        cube, gm, non_cube = decode_cube(dataset)
        dataset = encode_cube(cube, gm, non_cube)

    The returned data cube comprises all variables in *dataset*
    whose dimensions are ("time" , [...], y_dim_name, x_dim_name).
    Here y_dim_name and x_dim_name are determined by the
    :class:GridMapping derived from *dataset*.
    
    :param dataset: The dataset.
    :param normalize: Whether to normalize the *dataset*,
        before the cube subset is determined.
        If normalisation fails, the cube subset is created from *dataset*.
    :param force_copy: whether to create a copy of this dataset
        even if this dataset is identical to its cube subset.
    :param force_non_empty: whether the resulting cube
        must have at least one data variable.
        If True, a :class:DatasetIsNotACubeError may be raised.
    :param force_geographic: whether a geographic grid mapping
        is required.
        If True, a :class:DatasetIsNotACubeError may be raised.
    :return: A 3-tuple comprising the data cube subset of *dataset*
        the cube's grid mapping, and the remaining variables.
    :raise DatasetIsNotACubeError: If it is not possible to
        determine a data cube subset from *dataset*.
    """
    if normalize:
        try:
            dataset = normalize_dataset(dataset)
        except ValueError:
            pass
    try:
        grid_mapping = GridMapping.from_dataset(dataset, tolerance=1e-4)
    except ValueError as e:
        raise DatasetIsNotACubeError(f'Failed to detect grid mapping:'
                                     f' {e}') from e
    if force_geographic and not grid_mapping.crs.is_geographic:
        # We will need to overcome this soon!
        raise DatasetIsNotACubeError(f'Grid mapping must use geographic CRS,'
                                     f' but was {grid_mapping.crs.name!r}')

    x_dim_name, y_dim_name = grid_mapping.xy_dim_names
    time_name = 'time'

    cube_vars = set()
    dropped_vars = set()
    for var_name, var in dataset.data_vars.items():
        if var.ndim >= 3 \
                and var.dims[0] == time_name \
                and var.dims[-2] == y_dim_name \
                and var.dims[-1] == x_dim_name \
                and np.all(var.shape) \
                and var.shape[-2] > 1 \
                and var.shape[-1] > 1:
            cube_vars.add(var_name)
        else:
            dropped_vars.add(var_name)

    if force_non_empty and len(dropped_vars) == len(dataset.data_vars):
        # Or just return empty dataset?
        raise DatasetIsNotACubeError(f'No variables found with dimensions'
                                     f' ({time_name!r}, [...]'
                                     f' {y_dim_name!r}, {x_dim_name!r})'
                                     f' or dimension sizes too small')

    if not force_copy and not dropped_vars:
        # Pure cube!
        return dataset, grid_mapping, xr.Dataset()

    cube = dataset.drop_vars(dropped_vars).assign_attrs(dataset.attrs)

    return (cube,
            grid_mapping,
            dataset.drop_vars(cube_vars))


def _normalize_zonal_lat_lon(ds: xr.Dataset) -> xr.Dataset:
    """
    In case that the dataset only contains lat_centers and is a zonal mean dataset,
    the longitude dimension created and filled with the variable value of certain latitude.
    :param ds: some xarray dataset
    :return: a normalized xarray dataset
    """

    if 'latitude_centers' not in ds.coords or 'lon' in ds.coords:
        return ds

    ds_zonal = ds.copy()
    resolution = (ds.latitude_centers[1].values - ds.latitude_centers[0].values)
    ds_zonal = ds_zonal.assign_coords(
        lon=[i + (resolution / 2) for i in np.arange(-180.0, 180.0, resolution)])

    for var in ds_zonal.data_vars:
        if 'latitude_centers' in ds_zonal[var].dims:
            ds_zonal[var] = xr.concat([ds_zonal[var] for _ in ds_zonal.lon], 'lon')
            ds_zonal[var]['lon'] = ds_zonal.lon
            var_dims = ds_zonal[var].attrs.get('dimensions', [])
            lat_center_index = var_dims.index('latitude_centers')
            var_dims.remove('latitude_centers')
            var_dims.append('lat')
            var_dims.append('lon')
            var_chunk_sizes = ds_zonal[var].attrs.get('chunk_sizes', [])
            lat_chunk_size = var_chunk_sizes[lat_center_index]
            del var_chunk_sizes[lat_center_index]
            var_chunk_sizes.append(lat_chunk_size)
            var_chunk_sizes.append(ds_zonal.lon.size)
    ds_zonal = ds_zonal.rename_dims({'latitude_centers': 'lat'})
    ds_zonal = ds_zonal.drop_vars('latitude_centers')
    ds_zonal = ds_zonal.assign_coords(lat=ds.latitude_centers.values)
    ds_zonal = ds_zonal.transpose(..., 'lat', 'lon')

    has_lon_bnds = 'lon_bnds' in ds_zonal.coords or 'lon_bnds' in ds_zonal
    if not has_lon_bnds:
        lon_values = [[i - (resolution / 2), i + (resolution / 2)] for i in ds_zonal.lon.values]
        ds_zonal = ds_zonal.assign_coords(lon_bnds=xr.DataArray(lon_values, dims=['lon', 'bnds']))
    has_lat_bnds = 'lat_bnds' in ds_zonal.coords or 'lat_bnds' in ds_zonal
    if not has_lat_bnds:
        lat_values = [[i - (resolution / 2), i + (resolution / 2)] for i in ds_zonal.lat.values]
        ds_zonal = ds_zonal.assign_coords(lat_bnds=xr.DataArray(lat_values, dims=['lat', 'bnds']))

    ds_zonal.lon.attrs['bounds'] = 'lon_bnds'
    ds_zonal.lon.attrs['long_name'] = 'longitude'
    ds_zonal.lon.attrs['standard_name'] = 'longitude'
    ds_zonal.lon.attrs['units'] = 'degrees_east'

    ds_zonal.lat.attrs['bounds'] = 'lat_bnds'
    ds_zonal.lat.attrs['long_name'] = 'latitude'
    ds_zonal.lat.attrs['standard_name'] = 'latitude'
    ds_zonal.lat.attrs['units'] = 'degrees_north'

    return ds_zonal


def _normalize_lat_lon(ds: xr.Dataset) -> xr.Dataset:
    """
    Rename variables named 'longitude' or 'long' to 'lon', and 'latitude' to 'lon'.
    :param ds: some xarray dataset
    :return: a normalized xarray dataset, or the original one
    """
    lat_name = get_lat_dim_name_impl(ds)
    lon_name = get_lon_dim_name_impl(ds)

    name_dict = dict()
    if lat_name and 'lat' not in ds:
        name_dict[lat_name] = 'lat'

    if lon_name and 'lon' not in ds:
        name_dict[lon_name] = 'lon'

    if name_dict:
        ds = ds.rename(name_dict)

    return ds


def _normalize_lat_lon_2d(ds: xr.Dataset) -> xr.Dataset:
    """
    Detect 2D 'lat', 'lon' variables that span a equi-rectangular grid. Then:
    Drop original 'lat', 'lon' variables
    Rename original dimensions names of 'lat', 'lon' variables, usually ('y', 'x'), to
    ('lat', 'lon').
    Insert new 1D 'lat', 'lon' coordinate variables with dimensions 'lat' and 'lon', respectively.
    :param ds: some xarray dataset
    :return: a normalized xarray dataset, or the original one
    """
    if not ('lat' in ds and 'lon' in ds):
        return ds

    lat_var = ds['lat']
    lon_var = ds['lon']

    lat_dims = lat_var.dims
    lon_dims = lon_var.dims
    if lat_dims != lon_dims:
        return ds

    spatial_dims = lon_dims
    if len(spatial_dims) != 2:
        return ds

    x_dim_name = spatial_dims[-1]
    y_dim_name = spatial_dims[-2]

    lat_data_1 = lat_var[:, 0]
    lat_data_2 = lat_var[:, -1]
    lon_data_1 = lon_var[0, :]
    lon_data_2 = lon_var[-1, :]

    equal_lat = np.allclose(lat_data_1, lat_data_2, equal_nan=True)
    equal_lon = np.allclose(lon_data_1, lon_data_2, equal_nan=True)

    # Drop lat lon in any case. If note qual_lat and equal_lon subset_spatial_impl will
    # subsequently fail with a ValidationError

    ds = ds.drop_vars(['lon', 'lat'])

    if not (equal_lat and equal_lon):
        return ds

    ds = ds.rename({
        x_dim_name: 'lon',
        y_dim_name: 'lat',
    })

    ds = ds.assign_coords(lon=np.array(lon_data_1), lat=np.array(lat_data_1))

    return ds


def _normalize_lon_360(ds: xr.Dataset) -> xr.Dataset:
    """
    Fix the longitude of the given dataset ``ds`` so that it ranges from -180 to +180 degrees.

    :param ds: The dataset whose longitudes may be given in the range 0 to 360.
    :return: The fixed dataset or the original dataset.
    """

    if 'lon' not in ds.coords:
        return ds

    lon_var = ds.coords['lon']

    if len(lon_var.shape) != 1:
        return ds

    lon_size = lon_var.shape[0]
    if lon_size < 2:
        return ds

    lon_size_05 = lon_size // 2
    lon_values = lon_var.values
    if not np.any(lon_values[lon_size_05:] > 180.):
        return ds

    delta_lon = lon_values[1] - lon_values[0]

    var_names = [var_name for var_name in ds.data_vars]

    ds = ds.assign_coords(lon=xr.DataArray(np.linspace(-180. + 0.5 * delta_lon,
                                                       +180. - 0.5 * delta_lon,
                                                       lon_size),
                                           dims=ds['lon'].dims,
                                           attrs=dict(long_name='longitude',
                                                      standard_name='longitude',
                                                      units='degrees east')))

    ds = adjust_spatial_attrs(ds, True)

    new_vars = dict()
    for var_name in var_names:
        var = ds[var_name]
        if 'lon' in var.dims:
            new_var = var.roll(lon=lon_size_05, roll_coords=False)
            new_var.encoding.update(var.encoding)
            new_vars[var_name] = new_var

    return ds.assign(**new_vars)


def _reverse_decreasing_lat(ds: xr.Dataset) -> xr.Dataset:
    """
    In case the latitude decreases, invert it
    :param ds: some xarray dataset
    :return: a normalized xarray dataset
    """
    try:
        if _is_lat_decreasing(ds.lat):
            ds = ds.sel(lat=slice(None, None, -1))
    except AttributeError:
        # The dataset doesn't have 'lat', probably not geospatial
        pass
    except ValueError:
        # The dataset still has an ND 'lat' array
        pass
    return ds


def _normalize_jd2datetime(ds: xr.Dataset) -> xr.Dataset:
    """
    Convert the time dimension of the given dataset from Julian date to
    datetime.

    :param ds: Dataset on which to run conversion
    """

    try:
        time = ds.time
    except AttributeError:
        return ds

    try:
        units = time.units
    except AttributeError:
        units = None

    try:
        long_name = time.long_name
    except AttributeError:
        long_name = None

    if units:
        units = units.lower().strip()

    if long_name:
        units = long_name.lower().strip()

    units = units or long_name

    if not units or units != 'time in julian days':
        return ds

    ds = ds.copy()
    # Decode JD time
    # noinspection PyTypeChecker
    tuples = [jd2gcal(x, 0) for x in ds.time.values]
    # Replace JD time with datetime
    ds['time'] = [datetime(x[0], x[1], x[2]) for x in tuples]
    # Adjust attributes
    ds.time.attrs['long_name'] = 'time'
    ds.time.attrs['calendar'] = 'standard'

    return ds


def normalize_coord_vars(ds: xr.Dataset) -> xr.Dataset:
    """
    Turn potential coordinate variables from data variables into coordinate variables.

    Any data variable is considered a coordinate variable

    * whose name is its only dimension name;
    * whose number of dimensions is two and where the first dimension name is also a variable name
        and whose last dimension is named "bnds".

    :param ds: The dataset
    :return: The same dataset or a shallow copy with potential coordinate
             variables turned into coordinate variables.
    """

    if 'bnds' not in ds.dims:
        return ds

    coord_var_names = set()
    for data_var_name in ds.data_vars:
        data_var = ds.data_vars[data_var_name]
        if is_coord_var(ds, data_var):
            coord_var_names.add(data_var_name)

    if not coord_var_names:
        return ds

    old_ds = ds
    ds = old_ds.drop_vars(coord_var_names)
    ds = ds.assign_coords(**{bounds_var_name: old_ds[bounds_var_name]
                             for bounds_var_name in coord_var_names})

    return ds


def normalize_missing_time(ds: xr.Dataset) -> xr.Dataset:
    """
    Add a time coordinate variable and their associated bounds coordinate variables
    if either temporal CF attributes ``time_coverage_start`` and ``time_coverage_end``
    are given or time information can be extracted from the file name but the time dimension is
    missing.

    In case the time information is given by a variable called 't' instead of 'time', it will be
    renamed into 'time'.

    The new time coordinate variable will be named ``time`` with dimension ['time'] and shape [1].
    The time bounds coordinates variable will be named ``time_bnds``
    with dimensions ['time', 'bnds'] and shape [1,2].
    Both are of data type ``datetime64``.

    :param ds: Dataset to adjust
    :return: Adjusted dataset
    """
    time_coverage_start, time_coverage_end = _get_time_coverage_from_ds(ds)

    if not time_coverage_start and not time_coverage_end:
        # Can't do anything
        return ds

    time = _get_valid_time_coord(ds, 'time')
    if time is None:
        time = _get_valid_time_coord(ds, 't')
        if time is not None:
            ds = ds.rename_vars({"t": "time"})
            ds = ds.assign_coords(time=('time', ds.time.data))
            time = ds.time
    if time is not None:
        if not time.dims:
            ds = ds.drop_vars('time')
        elif len(time.dims) == 1:
            time_dim_name = time.dims[0]
            is_time_used_as_dim = any([(time_dim_name in ds[var_name].dims)
                                       for var_name in ds.data_vars])
            if is_time_used_as_dim:
                # It seems we already have valid time coordinates
                return ds
            time_bnds_var_name = time.attrs.get('bounds')
            if time_bnds_var_name in ds:
                ds = ds.drop_vars(time_bnds_var_name)
            ds = ds.drop_vars('time')
            ds = ds.drop_vars([var_name for var_name in ds.coords
                               if time_dim_name in ds.coords[var_name].dims])

    try:
        ds = ds.expand_dims('time')
    except BaseException as e:
        warnings.warn(f'failed to add time dimension: {e}')
    if time_coverage_start and time_coverage_end:
        time_value = time_coverage_start + 0.5 * (time_coverage_end - time_coverage_start)
    else:
        time_value = time_coverage_start or time_coverage_end
    new_coord_vars = dict(time=xr.DataArray([time_value], dims=['time']))
    if time_coverage_start and time_coverage_end:
        has_time_bnds = 'time_bnds' in ds.coords or 'time_bnds' in ds
        if not has_time_bnds:
            new_coord_vars.update(
                time_bnds=xr.DataArray([[time_coverage_start, time_coverage_end]],
                                       dims=['time', 'bnds']))
    ds = ds.assign_coords(**new_coord_vars)
    ds.coords['time'].attrs['long_name'] = 'time'
    ds.coords['time'].attrs['standard_name'] = 'time'
    ds.coords['time'].encoding['units'] = 'days since 1970-01-01'
    if 'time_bnds' in ds.coords:
        ds.coords['time'].attrs['bounds'] = 'time_bnds'
        ds.coords['time_bnds'].attrs['long_name'] = 'time'
        ds.coords['time_bnds'].attrs['standard_name'] = 'time'
        ds.coords['time_bnds'].encoding['units'] = 'days since 1970-01-01'

    return ds


def _get_valid_time_coord(ds: xr.Dataset, name: str) -> Optional[xr.DataArray]:
    if name in ds:
        time = ds[name]
        if time.size > 0 \
                and time.ndim == 1 \
                and _is_supported_time_dtype(time.dtype):
            return time
    return None


def _is_supported_time_dtype(dtype: np.dtype) -> bool:
    return any(np.issubdtype(dtype, t) for t in DatetimeTypes)


def _get_time_coverage_from_ds(ds: xr.Dataset) -> (pd.Timestamp, pd.Timestamp):
    time_coverage_start = ds.attrs.get('time_coverage_start')
    if time_coverage_start is not None:
        time_coverage_start = get_timestamp_from_string(time_coverage_start)

    time_coverage_end = ds.attrs.get('time_coverage_end')
    if time_coverage_end is not None:
        time_coverage_end = get_timestamp_from_string(time_coverage_end)

    if time_coverage_start or time_coverage_end:
        return time_coverage_start, time_coverage_end

    filename = ds.encoding.get('source', '').split('/')[-1]
    return get_timestamps_from_string(filename)


# TODO (forman): replace adjust_spatial_attrs function
#   by another on that uses a dataset's grid mapping,
#   see code in xcube/core/gen2/local/mladjuster.py

def adjust_spatial_attrs(ds: xr.Dataset, allow_point: bool = False) \
        -> xr.Dataset:
    """
    Adjust the global spatial attributes of the dataset by doing some
    introspection of the dataset and adjusting the appropriate attributes
    accordingly.

    In case the determined attributes do not exist in the dataset, these will
    be added.

    For more information on suggested global attributes see
    `Attribute Convention for Data Discovery
    <http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery>`_

    :param ds: Dataset to adjust
    :param allow_point: Whether to accept single point cells
    :return: Adjusted dataset
    """

    copied = False

    for dim in ('lon', 'lat'):
        geo_spatial_attrs = get_geo_spatial_attrs_from_var(
            ds, dim, allow_point=allow_point
        )
        if geo_spatial_attrs:
            # Copy any new attributes into the shallow Dataset copy
            for key in geo_spatial_attrs:
                if geo_spatial_attrs[key] is not None:
                    if not copied:
                        ds = ds.copy()
                        copied = True
                    ds.attrs[key] = geo_spatial_attrs[key]

    lon_min = ds.attrs.get('geospatial_lon_min')
    lat_min = ds.attrs.get('geospatial_lat_min')
    lon_max = ds.attrs.get('geospatial_lon_max')
    lat_max = ds.attrs.get('geospatial_lat_max')
    # TODO (forman): add geospatial_lon/lat_resolution

    if lon_min is not None \
            and lat_min is not None \
            and lon_max is not None \
            and lat_max is not None:

        if not copied:
            ds = ds.copy()

        ds.attrs['geospatial_bounds_crs'] = 'CRS84'
        ds.attrs['geospatial_bounds'] = f'POLYGON((' \
                                        f'{lon_min} {lat_min}, ' \
                                        f'{lon_min} {lat_max}, ' \
                                        f'{lon_max} {lat_max}, ' \
                                        f'{lon_max} {lat_min}, ' \
                                        f'{lon_min} {lat_min}))'

    return ds


def is_coord_var(ds: xr.Dataset, var: xr.DataArray):
    if len(var.dims) == 1 and var.name == var.dims[0]:
        return True
    return is_bounds_var(ds, var)


def is_bounds_var(ds: xr.Dataset, var: xr.DataArray):
    if len(var.dims) == 2 \
            and var.shape[1] == 2 \
            and var.dims[1] == 'bnds':
        coord_name = var.dims[0]
        return coord_name in ds
    return False


def get_geo_spatial_attrs_from_var(ds: xr.Dataset,
                                   var_name: str,
                                   allow_point: bool = False) -> Optional[dict]:
    """
    Get spatial boundaries, resolution and units of the given dimension of the given
    dataset. If the 'bounds' are explicitly defined, these will be used for
    boundary calculation, otherwise it will rest purely on information gathered
    from 'dim' itself.

    :param ds: The dataset
    :param var_name: The variable/dimension name.
    :param allow_point: True, if it is ok to have no actual spatial extent.
    :return: A dictionary {'attr_name': attr_value}
    """

    if var_name not in ds:
        return None

    var = ds[var_name]
    res_name = 'geospatial_{}_resolution'.format(var_name)
    min_name = 'geospatial_{}_min'.format(var_name)
    max_name = 'geospatial_{}_max'.format(var_name)
    units_name = 'geospatial_{}_units'.format(var_name)

    if 'bounds' in var.attrs:
        # According to CF Conventions the corresponding 'bounds' coordinate variable name
        # should be in the attributes of the coordinate variable
        bnds_name = var.attrs['bounds']
    else:
        # If 'bounds' attribute is missing, the bounds coordinate variable may be named
        # "<dim>_bnds"
        bnds_name = '%s_bnds' % var_name

    dim_var = None

    if bnds_name in ds:
        bnds_var = ds[bnds_name]
        if len(bnds_var.shape) == 2 and bnds_var.shape[0] > 0 and bnds_var.shape[1] == 2:
            dim_var = bnds_var
            dim_res = abs(bnds_var[0, 1] - bnds_var[0, 0])
            if bnds_var.shape[0] > 1:
                dim_min = min(bnds_var[0, 0], bnds_var[-1, 1])
                dim_max = max(bnds_var[0, 0], bnds_var[-1, 1])
            else:
                dim_min = min(bnds_var[0, 0], bnds_var[0, 1])
                dim_max = max(bnds_var[0, 0], bnds_var[0, 1])

    if dim_var is None:
        if len(var.shape) == 1 and var.shape[0] > 0:
            if var.shape[0] > 1:
                dim_var = var
                dim_res = abs(var[1] - var[0])
                dim_min = min(var[0], var[-1]) - 0.5 * dim_res
                dim_max = max(var[0], var[-1]) + 0.5 * dim_res
            elif var.size == 1:
                if res_name in ds.attrs:
                    dim_res = ds.attrs[res_name]
                    if isinstance(dim_res, str):
                        # remove any units from string
                        dim_res = dim_res.replace('degree', '').replace('deg', ''). \
                            replace('°', '').strip()
                    try:
                        dim_res = float(dim_res)
                        dim_var = var
                        # Consider extent in metadata if provided
                        dim_min = var[0] - 0.5 * dim_res
                        dim_max = var[0] + 0.5 * dim_res
                    except ValueError:
                        if allow_point:
                            dim_var = var
                            # Actually a point with no extent
                            dim_res = 0.0
                            dim_min = var[0]
                            dim_max = var[0]
                elif allow_point:
                    dim_var = var
                    # Actually a point with no extent
                    dim_res = 0.0
                    dim_min = var[0]
                    dim_max = var[0]

    if dim_var is None:
        # Cannot determine spatial extent for variable/dimension var_name
        return None

    if 'units' in var.attrs:
        dim_units = var.attrs['units']
    else:
        dim_units = None

    geo_spatial_attrs = dict()
    # noinspection PyUnboundLocalVariable
    geo_spatial_attrs[res_name] = float(dim_res)
    # noinspection PyUnboundLocalVariable
    geo_spatial_attrs[min_name] = float(dim_min)
    # noinspection PyUnboundLocalVariable
    geo_spatial_attrs[max_name] = float(dim_max)
    geo_spatial_attrs[units_name] = dim_units

    return geo_spatial_attrs


def get_lon_dim_name_impl(ds: Union[xr.Dataset, xr.DataArray]) -> Optional[str]:
    """
    Get the name of the longitude dimension.
    :param ds: An xarray Dataset
    :return: the name or None
    """
    return _get_dim_name(ds, ['lon', 'longitude', 'long'])


def get_lat_dim_name_impl(ds: Union[xr.Dataset, xr.DataArray]) -> Optional[str]:
    """
    Get the name of the latitude dimension.
    :param ds: An xarray Dataset
    :return: the name or None
    """
    return _get_dim_name(ds, ['lat', 'latitude'])


def _get_dim_name(ds: Union[xr.Dataset, xr.DataArray], possible_names: Sequence[str]) \
        -> Optional[str]:
    for name in possible_names:
        if name in ds.dims:
            return name
    return None


def _is_lat_decreasing(lat: xr.DataArray) -> bool:
    """
    Determine if the latitude is decreasing
    """
    if lat[0] > lat[-1]:
        return True

    return False


def _normalize_dim_order(ds: xr.Dataset) -> xr.Dataset:
    copy_created = False

    for var_name in ds.data_vars:
        var = ds[var_name]
        dim_names = list(var.dims)
        num_dims = len(dim_names)
        if num_dims == 0:
            continue

        must_transpose = False

        if 'time' in dim_names:
            time_index = dim_names.index('time')
            if time_index > 0:
                must_transpose = _swap_pos(dim_names, time_index, 0)

        if num_dims >= 2 and 'lat' in dim_names and 'lon' in dim_names:
            lat_index = dim_names.index('lat')
            if lat_index != num_dims - 2:
                must_transpose = _swap_pos(dim_names, lat_index, -2)
            lon_index = dim_names.index('lon')
            if lon_index != num_dims - 1:
                must_transpose = _swap_pos(dim_names, lon_index, -1)

        if must_transpose:
            if not copy_created:
                ds = ds.copy()
                copy_created = True
            ds[var_name] = var.transpose(*dim_names)
            if var.encoding and 'chunksize' in ds[var_name].data:
                ds[var_name].encoding['chunks'] = ds[var_name].data.chunksize

    return ds


def _swap_pos(lst, i1, i2):
    e1, e2 = lst[i1], lst[i2]
    lst[i2], lst[i1] = e1, e2
    return True



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Dict, Union, List, Any

from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from ..response import CubeGeneratorResult
from ..response import CubeInfo
from ..response import make_cube_generator_result_class


class CubeGeneratorToken(JsonObject):
    def __init__(self, access_token: str, token_type: str):
        self.access_token = access_token
        self.token_type = token_type

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                access_token=JsonStringSchema(min_length=1),
                token_type=JsonStringSchema(min_length=1)
            ),
            required=['access_token', 'token_type'],
            additional_properties=False,
            factory=cls
        )

    @classmethod
    def from_dict(cls, value: Dict) -> 'CubeGeneratorToken':
        return cls.get_schema().from_instance(value)


class CubeGeneratorJobStatus(JsonObject):
    # noinspection PyUnusedLocal
    def __init__(self,
                 succeeded: int = None,
                 failed: int = None,
                 active: int = None,
                 start_time: str = None,
                 completion_time: str = None,
                 conditions: List[Dict[str, Any]] = None,
                 **additional_properties):
        self.succeeded: Optional[int] = succeeded
        self.failed: Optional[int] = failed
        self.active: Optional[int] = active
        self.start_time: Optional[str] = start_time
        self.completion_time: Optional[str] = completion_time
        self.conditions: Optional[Dict[str, Any]] = conditions
        self.additional_properties: Dict[str, Any] = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                succeeded=JsonIntegerSchema(nullable=True),
                failed=JsonIntegerSchema(nullable=True),
                active=JsonIntegerSchema(nullable=True),
                start_time=JsonStringSchema(nullable=True),
                completion_time=JsonStringSchema(nullable=True),
                conditions=JsonArraySchema(
                    items=JsonObjectSchema(additional_properties=True),
                    nullable=True
                )
            ),
            additional_properties=True,
            factory=cls
        )

    @classmethod
    def from_dict(cls, value: Dict) -> 'CubeGeneratorJobStatus':
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgressState(JsonObject):
    """Current progress state of the remote generator."""

    def __init__(self,
                 progress: float,
                 # worked: Union[int, float],
                 total_work: Union[int, float],
                 **additional_properties):
        self.progress: float = float(progress)
        # self.worked: float = float(worked)
        self.total_work: float = float(total_work)
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                progress=JsonNumberSchema(minimum=0.0),
                # worked=JsonNumberSchema(minimum=0),
                total_work=JsonNumberSchema(exclusive_minimum=0)
            ),
            required=[
                'progress',
                # 'worked',
                'total_work',
            ],
            additional_properties=True,
            factory=cls)

    @classmethod
    def from_dict(cls, value: Dict) -> 'CubeGeneratorProgressState':
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgress(JsonObject):
    """Current progress of the remote generator."""

    def __init__(self,
                 sender: str,
                 state: CubeGeneratorProgressState):
        self.sender: str = sender
        self.state: CubeGeneratorProgressState = state

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                sender=JsonStringSchema(),
                state=CubeGeneratorProgressState.get_schema(),
            ),
            required=[
                'sender',
                'state',
            ],
            additional_properties=False,
            factory=cls)

    @classmethod
    def from_dict(cls, value: Dict) -> 'CubeGeneratorProgress':
        return cls.get_schema().from_instance(value)


class CubeGeneratorState(JsonObject):
    """Current state of the remote generator."""

    def __init__(self,
                 job_id: str,
                 job_status: CubeGeneratorJobStatus,
                 job_result: Optional[CubeGeneratorResult] = None,
                 output: Optional[List[str]] = None,
                 progress: Optional[List[CubeGeneratorProgress]] = None,
                 **additional_properties):
        self.job_id = job_id
        self.job_status = job_status
        self.job_result = job_result
        self.output = output
        self.progress = progress
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                job_id=JsonStringSchema(min_length=1),
                job_status=CubeGeneratorJobStatus.get_schema(),
                job_result=CubeGeneratorResult.get_schema(),
                output=JsonArraySchema(
                    items=JsonStringSchema(),
                    nullable=True
                ),
                progress=JsonArraySchema(
                    items=CubeGeneratorProgress.get_schema(),
                    nullable=True
                )
            ),
            required=['job_id', 'job_status'],
            additional_properties=True,
            factory=cls
        )

    @classmethod
    def from_dict(cls, value: Dict) -> 'CubeGeneratorState':
        return cls.get_schema().from_instance(value)


class CostEstimation(JsonObject):
    def __init__(self,
                 required: int,
                 available: int = None,
                 limit: int = None):
        self.available: int = available
        self.limit: Optional[int] = limit
        self.required: Optional[int] = required

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                required=JsonIntegerSchema(minimum=0),
                available=JsonIntegerSchema(minimum=0),
                limit=JsonIntegerSchema(minimum=0),
            ),
            required=['required'],
            additional_properties=False,
            factory=cls)

    @classmethod
    def from_dict(cls, value: Dict) -> 'CostEstimation':
        return cls.get_schema().from_instance(value)


class CubeInfoWithCosts(CubeInfo):
    def __init__(self, cost_estimation: CostEstimation, **kwargs):
        super().__init__(**kwargs)
        self.cost_estimation: CostEstimation = cost_estimation

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(cost_estimation=CostEstimation.get_schema())
        schema.required.append('cost_estimation')
        schema.factory = cls
        return schema


CubeInfoWithCostsResult = make_cube_generator_result_class(CubeInfoWithCosts)
